# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""utils for tinyms train, eval and predict"""
import random
import logging
import os
import time
from datetime import datetime
import numpy as np
from tinyms import Tensor
from mindspore.train.serialization import load_checkpoint, load_param_into_net, save_checkpoint
from tinyms.data.utils import save_image


__all__ = [
    'GanReporter',
    'gan_load_ckpt',
    'GanImagePool',
]


class GanReporter(logging.Logger):
    """
    Reporter class for Cycle Gan.
    This class includes several functions that can save images/checkpoints and print/save logging information.

    Args:
        args (class): Option class.
    """
    def __init__(self, args):
        super(GanReporter, self).__init__("cyclegan")
        self.log_dir = os.path.join(args.outputs_dir, 'log')
        self.imgs_dir = os.path.join(args.outputs_dir, 'imgs')
        self.ckpts_dir = os.path.join(args.outputs_dir, 'ckpt')
        if not os.path.exists(self.log_dir):
            os.makedirs(self.log_dir, exist_ok=True)
        if not os.path.exists(self.imgs_dir):
            os.makedirs(self.imgs_dir, exist_ok=True)
        if not os.path.exists(self.ckpts_dir):
            os.makedirs(self.ckpts_dir, exist_ok=True)
        self.rank = 0
        self.save_checkpoint_epochs = args.save_checkpoint_epochs
        self.save_imgs = args.save_imgs
        # console handler
        console = logging.StreamHandler()
        console.setLevel(logging.INFO)
        formatter = logging.Formatter('%(message)s')
        console.setFormatter(formatter)
        self.addHandler(console)
        # file handler
        log_name = datetime.now().strftime('%Y-%m-%d_time_%H_%M_%S') + '_rank_{}.log'.format(self.rank)
        self.log_fn = os.path.join(self.log_dir, log_name)
        fh = logging.FileHandler(self.log_fn)
        fh.setLevel(logging.INFO)
        fh.setFormatter(formatter)
        self.addHandler(fh)
        self.save_args(args)
        self.step = 0
        self.epoch = 0
        self.dataset_size = args.dataset_size
        self.print_iter = 100
        self.G_loss = []
        self.D_loss = []

    def info(self, msg, *args, **kwargs):
        if self.isEnabledFor(logging.INFO):
            self._log(logging.INFO, msg, args, **kwargs)

    def save_args(self, args):
        """Show args configuration info."""
        self.info('Args:')
        args_dict = vars(args)
        for key in args_dict.keys():
            self.info('--> %s: %s', key, args_dict[key])
        self.info('')

    def important_info(self, msg, *args, **kwargs):
        if self.logger.isEnabledFor(logging.INFO) and self.rank == 0:
            line_width = 2
            important_msg = '\n'
            important_msg += ('*'*70 + '\n')*line_width
            important_msg += ('*'*line_width + '\n')*2
            important_msg += '*'*line_width + ' '*8 + msg + '\n'
            important_msg += ('*'*line_width + '\n')*2
            important_msg += ('*'*70 + '\n')*line_width
            self.info(important_msg, *args, **kwargs)

    def epoch_start(self):
        """Print log when step start."""
        self.step_start_time = time.time()
        self.epoch_start_time = time.time()
        self.step = 0
        self.epoch += 1
        self.G_loss = []
        self.D_loss = []

    def step_end(self, res_G, res_D):
        """
        Print log when step end.

        Args:
            res_G (layers.Layer): TrainOneStepG instance.
            res_D (layers.Layer): TrainOneStepD instance.
        """
        self.step += 1
        loss_D = float(res_D.asnumpy())
        res = []
        for item in res_G[2:]:
            res.append(float(item.asnumpy()))
        self.G_loss.append(res[0])
        self.D_loss.append(loss_D)
        if self.step % self.print_iter == 0:
            step_cost = (time.time() - self.step_start_time) * 1000 / self.print_iter
            losses = "G_loss: {:.2f}, D_loss:{:.2f}, loss_G_A: {:.2f}, loss_G_B: {:.2f}, loss_C_A: {:.2f},"\
                     "loss_C_B: {:.2f}, loss_idt_A: {:.2f}, loss_idt_B：{:.2f}".format(
                         res[0], loss_D, res[1], res[2], res[3], res[4], res[5], res[6])
            self.info("Epoch[{}] [{}/{}] step cost: {:.2f} ms, {}".format(
                self.epoch, self.step, self.dataset_size, step_cost, losses))
            self.step_start_time = time.time()

    def epoch_end(self, net):
        """
        Print log and save checkpoints when epoch end.

        Args:
            net (layers.Layer): TrainOneStepG instance.
        """
        epoch_cost = (time.time() - self.epoch_start_time) * 1000
        pre_step_time = epoch_cost / self.dataset_size
        mean_loss_G = sum(self.G_loss) / self.dataset_size
        mean_loss_D = sum(self.D_loss) / self.dataset_size
        self.info("Epoch [{}] total cost: {:.2f} ms, pre step: {:.2f} ms, G_loss: {:.2f}, D_loss: {:.2f}".format(
            self.epoch, epoch_cost, pre_step_time, mean_loss_G, mean_loss_D))

        if self.epoch % self.save_checkpoint_epochs == 0 and self.rank == 0:
            save_checkpoint(net.G.generator.G_A, os.path.join(self.ckpts_dir, f"G_A_{self.epoch}.ckpt"))
            save_checkpoint(net.G.generator.G_B, os.path.join(self.ckpts_dir, f"G_B_{self.epoch}.ckpt"))
            save_checkpoint(net.G.D_A, os.path.join(self.ckpts_dir, f"D_A_{self.epoch}.ckpt"))
            save_checkpoint(net.G.D_B, os.path.join(self.ckpts_dir, f"D_B_{self.epoch}.ckpt"))

    def visualizer(self, img_A, img_B, fake_A, fake_B):
        """
        Save visualized image.

        Args:
            img_A (numpy.ndarray): Image data.
            img_B (numpy.ndarray): Image data.
            fake_A (numpy.ndarray): Generated image data.
            fake_B (numpy.ndarray): Generated image data.
        """
        if self.save_imgs and self.step % self.dataset_size == 0 and self.rank == 0:
            save_image(img_A, os.path.join(self.imgs_dir, f"{self.epoch}_img_A.jpg"))
            save_image(img_B, os.path.join(self.imgs_dir, f"{self.epoch}_img_B.jpg"))
            save_image(fake_A, os.path.join(self.imgs_dir, f"{self.epoch}_fake_A.jpg"))
            save_image(fake_B, os.path.join(self.imgs_dir, f"{self.epoch}_fake_B.jpg"))

    def start_predict(self, direction):
        """
        Print log when predict start.

        Args:
            direction (str): The predict name.
        """
        self.predict_start_time = time.time()
        self.direction = direction
        self.info('==========start predict %s===============', self.direction)

    def end_predict(self):
        """Print log when predict end."""
        cost = (time.time() - self.predict_start_time) * 1000
        pre_step_cost = cost / self.dataset_size
        self.info('total {} imgs cost {:.2f} ms, pre img cost {:.2f}'.format(self.dataset_size, cost, pre_step_cost))
        self.info('==========end predict %s===============\n', self.direction)

    def start_eval(self):
        """Print log when eval start."""
        self.eval_start_time = time.time()
        self.info('==========start eval %s===============')

    def end_eval(self):
        """Print log when eval end."""
        cost = (time.time() - self.eval_start_time) * 1000
        pre_step_cost = cost / self.dataset_size
        self.info('total {} imgs cost {:.2f} ms, pre img cost {:.2f}'.format(self.dataset_size, cost, pre_step_cost))
        self.info('==========end eval %s===============\n')


class GanImagePool():
    """
    This class implements an image buffer that stores previously generated images.

    This buffer enables us to update discriminators using a history of generated images
    rather than the ones produced by the latest generators.

    Args:
        pool_size (int): The size of image buffer, if pool_size=0, no buffer will be created.
    """

    def __init__(self, pool_size):
        """Initialize the ImagePool class."""
        self.pool_size = pool_size
        if self.pool_size > 0:  # create an empty pool
            self.num_imgs = 0
            self.images = []

    def query(self, images):
        """
        Query an image from the pool.

        By 50/100, the buffer will return input images.
        By 50/100, the buffer will return images previously stored in the buffer,
        and insert the current images to the buffer.

        Args:
            images (Tensor): The latest generated images from the generator

        Returns:
            Images tensor from the buffer.
        """
        if isinstance(images, Tensor):
            images = images.asnumpy()
        if self.pool_size == 0:  # if the buffer size is 0, do nothing
            return Tensor(images)
        return_images = []
        for image in images:
            # if the buffer is not full; keep inserting current images to the buffer
            if self.num_imgs < self.pool_size:
                self.num_imgs = self.num_imgs + 1
                self.images.append(image)
                return_images.append(image)
            else:
                p = random.uniform(0, 1)
                # by 50% chance, the buffer will return a previously stored image
                # and insert the current image into the buffer
                if p > 0.5:
                    random_id = random.randint(0, self.pool_size - 1)  # randint is inclusive
                    tmp = self.images[random_id].copy()
                    self.images[random_id] = image
                    return_images.append(tmp)
                else:       # by another 50% chance, the buffer will return the current image
                    return_images.append(image)
        return_images = np.array(return_images)   # collect all the images and return
        if len(return_images.shape) != 4:
            raise ValueError("img should be 4d, but get shape {}".format(return_images.shape))
        return Tensor(return_images)


def gan_load_ckpt(G_A_ckpt=None, G_B_ckpt=None, D_A_ckpt=None, D_B_ckpt=None,
                  G_A=None, G_B=None, D_A=None, D_B=None):
    """
    Load parameter from checkpoint files.

    Args:
        G_A_ckpt (Checkpoint): Load G_A checkpoint file.
        G_B_ckpt (Checkpoint): Load G_B checkpoint file.
        D_A_ckpt (Checkpoint): Load D_A checkpoint file.
        D_B_ckpt (Checkpoint): Load D_B checkpoint file.
        G_A (Generator): G_A Generator.
        G_B (Generator): G_B Generator.
        D_A (Discriminator): D_A Discriminator.
        D_B (Discriminator): D_B Discriminator.
    """
    if G_A_ckpt is not None:
        param_GA = load_checkpoint(G_A_ckpt)
        load_param_into_net(G_A, param_GA)
    if G_B_ckpt is not None:
        param_GB = load_checkpoint(G_B_ckpt)
        load_param_into_net(G_B, param_GB)
    if D_A is not None and D_A_ckpt is not None:
        param_DA = load_checkpoint(D_A_ckpt)
        load_param_into_net(D_A, param_DA)
    if D_B is not None and D_B_ckpt is not None:
        param_DB = load_checkpoint(D_B_ckpt)
        load_param_into_net(D_B, param_DB)



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import numpy as np
from scipy.stats import truncnorm

import tinyms as ts
from tinyms import layers, Tensor
from tinyms.layers import ReLU, MaxPool2d, Flatten, Dropout


def _conv_variance_scaling_initializer(in_channel, out_channel, kernel_size):
    fan_in = in_channel * kernel_size * kernel_size
    scale = 1.0
    scale /= max(1., fan_in)
    stddev = (scale ** 0.5) / .87962566103423978
    mu, sigma = 0, stddev
    weight = truncnorm(-2, 2, loc=mu, scale=sigma).rvs(out_channel * in_channel * kernel_size * kernel_size)
    return ts.reshape(weight, (out_channel, in_channel, kernel_size, kernel_size))


def _weight_variable(shape, factor=0.01):
    init_value = np.random.randn(*shape).astype(np.float32) * factor
    return Tensor(init_value)


def _conv3x3(in_channel, out_channel, stride=1):
    weight_shape = (out_channel, in_channel, 3, 3)
    weight = _weight_variable(weight_shape)
    return layers.Conv2d(in_channel, out_channel,
                         kernel_size=3, stride=stride, padding=0, pad_mode='same', weight_init=weight)


def _conv1x1(in_channel, out_channel, stride=1):
    weight_shape = (out_channel, in_channel, 1, 1)
    weight = _weight_variable(weight_shape)
    return layers.Conv2d(in_channel, out_channel,
                         kernel_size=1, stride=stride, padding=0, pad_mode='same', weight_init=weight)


def _conv7x7(in_channel, out_channel, stride=1):
    weight_shape = (out_channel, in_channel, 7, 7)
    weight = _weight_variable(weight_shape)
    return layers.Conv2d(in_channel, out_channel,
                         kernel_size=7, stride=stride, padding=0, pad_mode='same', weight_init=weight)


def _conv11x11(in_channel, out_channel, stride=1):
    weight_shape = (out_channel, in_channel, 11, 11)
    weight = _weight_variable(weight_shape)
    return layers.Conv2d(in_channel, out_channel,
                         kernel_size=11, stride=stride, padding=2, pad_mode='pad', weight_init=weight)


def _conv5x5(in_channel, out_channel, stride=1):
    weight_shape = (out_channel, in_channel, 5, 5)
    weight = _weight_variable(weight_shape)
    return layers.Conv2d(in_channel, out_channel,
                         kernel_size=5, stride=stride, padding=2, pad_mode='pad', weight_init=weight)


def _bn(channel):
    return layers.BatchNorm2d(channel, eps=1e-4, momentum=0.9,
                              gamma_init=1, beta_init=0, moving_mean_init=0, moving_var_init=1)


def _bn_last(channel):
    return layers.BatchNorm2d(channel, eps=1e-4, momentum=0.9,
                              gamma_init=0, beta_init=0, moving_mean_init=0, moving_var_init=1)


def _fc(in_channel, out_channel):
    weight_shape = (out_channel, in_channel)
    weight = _weight_variable(weight_shape)
    return layers.Dense(in_channel, out_channel, has_bias=True, weight_init=weight, bias_init=0)


class AlexNet(layers.Layer):
    """
    Get AlexNet neural network.

    Args:
        class_num (int): Class number. Default: 1000.

    Returns:
        layers.Layer, layer instance of AlexNet neural network.

    Examples:
        >>> from tinyms.model import AlexNet
        >>>
        >>> net = AlexNet(class_num=1000)
    """

    def __init__(self, class_num=1000):
        super(AlexNet, self).__init__()

        self.features = layers.SequentialLayer(
            [
                _conv11x11(3, 64, 4),
                ReLU(),
                MaxPool2d(kernel_size=3, stride=2),
                _conv5x5(64, 192),
                ReLU(),
                MaxPool2d(kernel_size=3, stride=2),
                _conv3x3(192, 384),
                ReLU(),
                _conv3x3(384, 256),
                ReLU(),
                _conv3x3(256, 256),
                ReLU(),
                MaxPool2d(kernel_size=3, stride=2),
                Flatten(),
                Dropout(),
                _fc(256*6*6, 4096),
                ReLU(),
                Dropout(),
                _fc(4096, 4096),
                ReLU(),
                _fc(4096, class_num)
            ]

        )

    def construct(self, x):
        x = self.features(x)
        return x


def alexnet(**kwargs):
    """
    Get AlexNet neural network.

    Args:
        class_num (int): Class number. Default: 10.

    Returns:
        layers.Layer, layer instance of AlexNet neural network.

    Examples:
        >>> from tinyms.model import alexnet
        >>>
        >>> net = alexnet(class_num=10)
    """
    return AlexNet(class_num=kwargs.get('class_num', 10))



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
import os
import yaml

__all__ = ['get_specified_config']


def _get_config_path():
    cur_path = os.path.dirname(__file__)
    return os.path.join(cur_path, "configs")


def _get_all_config(yaml_file):
    if not os.path.exists(yaml_file):
        raise FileNotFoundError('The transform yaml_file config file not found!')

    all_transform_configs = []
    with open(yaml_file, mode='r', encoding='utf-8') as f:
        for transform_config in yaml.safe_load_all(f):
            all_transform_configs.append(transform_config)
    return all_transform_configs


def get_specified_config(transforms_op, yaml_path=None):
    r'''
    Get specified vision transform parameters from transform yaml file.

    Args:
        transforms_op (str): Specified vision transforms class, such as: VOCTransform, MnistTransform.
        yaml_path (str): The yaml file path of the vision transform configuration. Default: None.

    Returns:
        dict, the vision transform parameters of the specified transforms_op.
    '''
    if not yaml_path:
        yaml_path = os.path.join(_get_config_path(), 'transform_config.yaml')
    all_transform_configs = _get_all_config(yaml_path)
    data = None
    for transform_config in all_transform_configs:
        if transforms_op in transform_config:
            data = transform_config[transforms_op]
    return data



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
import base64
import os
import sys
import json
import signal
import socket
import logging
import platform
import subprocess
import cv2
import numpy as np

from flask import request, Flask, jsonify, make_response
from flask_cors import CORS, cross_origin
from tinyms.serving.servable import predict, servable_search, web_predict

app = Flask(__name__)
app.config['MAX_CONTENT_LENGTH'] = 20*1024*1024
CORS(app, supports_credentials=True)


def set_cross(res):
    response = make_response(jsonify(res))
    response.headers['Access-Control-Allow-Origin'] = '*'
    response.headers['Access-Control-Allow-Method'] = '*'
    response.headers['Access-Control-Allow-Headers'] = '*'


def base64tonumpy(img_base64):
    np_arr = np.fromstring(img_base64, np.uint8)
    img = cv2.imdecode(np_arr, cv2.IMREAD_UNCHANGED)
    return np.array(img)


@app.route('/servables', methods=['GET'])
def list_servables():
    """
    Handle the list_servables request sent by client, call the servable search in tinyms.serving.servable

    Whether the server started or not will be detected first.

    Returns:
        A json object of servable information in the backend will be sent back to the client.

    Examples:
        >>> # In the client part, the servable search request will be routed and processed here
        >>> res = requests.get(url=url, headers=headers)
        >>> res_body = res.json()
    """

    return jsonify(servable_search())


@app.route('/servables/<name>', methods=['POST'])
def predict_server(name):
    """
    Handle the request sent by client, call the servable_search, predict function
    in tinyms.serving.servable and return the json result to the client.

    Whether the server started or not will be detected first.

    Args:
        name (str): Specifies the servable name by client.

    Returns:
        A json object of predicted result will be sent back to the client.
    """

    json_data = request.get_json()
    instance = json_data['instance']
    strategy = json_data['strategy']

    res = servable_search(name)
    if res['status'] != 0:
        return jsonify(res)
    servable = res['servables'][0]
    res = predict(instance, name, servable['model'], strategy)
    return jsonify(res)


@app.route('/predict', methods=['POST'])
@cross_origin(supports_credentials=True)
def predict_web():
    json_data = json.loads(request.data)
    strategy = json_data['strategy']
    servable_name = json_data['servable_name']
    dataset_name = servable_name.split('_')[-1]
    data = base64.b64decode(json_data['data'])
    image = base64tonumpy(data)

    instance = {
        'shape': list(image.shape),
        'dtype': image.dtype.name,
        'data': json.dumps(image.tolist())
    }
    res = servable_search(servable_name)
    if res['status'] != 0:
        set_cross(res)
        return jsonify(res)
    servable = res['servables'][0]
    res = web_predict(instance, servable_name, servable['model'], dataset_name, strategy)
    set_cross(res)
    return jsonify(res)


class _FlaskServer(object):
    """
    Create a flask service, only be used to trigger starting the flask server
    in subprocess.
    """

    def __init__(self, host='127.0.0.1', port=5000):
        self.system_name = platform.system().lower()
        self.host = host
        self.port = port

    def signal_handler(self, signal, frame):
        self.shutdown()
        sys.exit(0)

    def shutdown(self):
        """
        Shutdown the flask server.
        """

        if self.system_name == "windows":
            server_res = subprocess.getoutput("netstat -ano | findstr 5000")
            server_pid = None
            for line in server_res.split("\n"):
                temp = [i for i in line.split(' ') if i != '']
                if len(temp) > 4:
                    if temp[1] == f'{self.host}:{self.port}' and temp[3] == 'LISTENING':
                        server_pid = temp[4]
                        continue
            if server_pid:
                os.system("taskkill /t /f /pid %s" % server_pid)
        else:
            server_pid = subprocess.getoutput("netstat -anp | grep %s | awk '{printf $7}' | cut -d/ -f1" % self.port)
            subprocess.run("kill -9 " + str(server_pid), shell=True)
        return 'Server shutting down...'

    def windows_run_server(self):
        cmd = 'python -c "from tinyms.serving import run_flask; run_flask()"'
        server_process = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)

        for sig in [signal.SIGINT, signal.SIGTERM]:
            signal.signal(sig, self.signal_handler)

    def linux_run_server(self):
        cmd = [f'python -c "from tinyms.serving import run_flask; run_flask()"']
        server_process = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
        for line in iter(server_process.stdout.readline, b''):
            print(line.decode('utf-8'))

        for sig in [signal.SIGINT, signal.SIGHUP, signal.SIGTERM]:
            signal.signal(sig, self.signal_handler)

    def run(self):
        if self.system_name == "windows":
            self.windows_run_server()
        else:
            self.linux_run_server()


def run_flask(host='127.0.0.1', port=5000):
    """
    Start the flask server, only be used to trigger starting the flask server in subprocess.

    Directly calling this function is not recommended, please use start_server(). Only Error message will be displayed.

    Args:
        host (str): the ip address of the flask server. Default: '127.0.0.1'.
        port (int): the port of the server. Default: 5000.

    Returns:
        Server Started

    Examples:
        >>> # In the start_server function
        >>> cmd = ['python -c "from tinyms.serving import run_flask; run_flask()"']
        >>> server_process = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
    """

    if os.path.exists('temp.json'):
        with open('temp.json', 'r') as f:
            config_data = json.load(f)
            host = config_data.get('host')
            port = config_data.get('port')

    log = logging.getLogger('werkzeug')
    log.setLevel(logging.ERROR)
    app.run(host=host, port=port)


class Server:
    '''
    Server is the entrance to initialize and shutdown the serving server, and
    also accept all requests from the client side by calling Flask server.

    Args:
        host (str): Serving server host ip. Default: '127.0.0.1'.
        port (int): Serving server listen port. Default: 5000.
        serving_path (str, optional): Set the read path of a service configuration.
            Default: '/etc/tinyms/serving/'.

    Examples:
        >>> from tinyms.serving import Server
        >>>
        >>> server = Server()
    '''

    def __init__(self, host='127.0.0.1', port=5000, serving_path='/etc/tinyms/serving/'):
        json_data = {}
        json_data.update({
            'host': host,
            'port': port,
            'serving_path': serving_path
        })
        with open('temp.json', 'w') as f:
            json.dump(json_data, f)

        self.host = host
        self.port = port

    def _check_started(self):
        """
        Detect whether the serving server is started or not.

        A bool value of True will be returned if the server is started, else False.

        Returns:
            A bool value of True (if server started) or False (if server not started).
        """
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            s.connect((self.host, self.port))
            s.shutdown(2)
            return True
        except:
            return False

    def start_server(self):
        """
        Start the flask server in a subprocess.

        Catch the signal of CTRL + D to shutdown, otherwise call shutdown() function
        to shutdown the server, if the ip and port already in use, server won't start
        for a second time.

        Returns:
            Start the server in a sub process.

        Examples:
            >>> from tinyms.serving import Server
            >>>
            >>> server = Server()
            >>> server.start_server()
            Server starts at host 127.0.0.1, port 5000
        """

        if self._check_started() is True:
            print('Server already started at host %s, port %d' % (self.host, self.port))
        else:
            # TODO: Add dynamic host ip and port support
            _FlaskServer(self.host, self.port).run()
            print('Server started at host %s, port %d' % (self.host, self.port))

    def shutdown(self):
        """
        Shutdown the flask server.

        Search fot the pid of the process running on port 5000, and kill it. This function
        will be automatically called when SIGINT, SIGHUP and SIGTERM signals caught.

        Returns:
            A string message of server shutting down or not.

        Examples:
            >>> from tinyms.serving import Server
            >>>
            >>> server = Server()
            >>> server.shutdown()
            Server shutting down...
        """
        if not self._check_started() is True:
            print('Server already shutdown at host %s, port %d' % (self.host, self.port))
        else:
            if os.path.exists('temp.json'):
                os.remove('temp.json')
            _FlaskServer(self.host, self.port).shutdown()



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
import os
import cv2
import numpy as np

import tinyms as ts
from tinyms import model
from tinyms.vision import voc_transform, shanshui_tranform

__all__ = ['ObjectDetector', 'object_detection_predict']

model_checker = {
    "ssd300": model.ssd300_mobilenetv2
}

transform_checker = {
    'voc': voc_transform,
    'shanshui': shanshui_tranform
}


class ObjectDetector():
    r"""
    ObjectDetector is a high-level class defined for building model，preproceing the input image,
    predicting and postprocessing the prediction output data.

    Args:
        config (dict): model config parsed from the json file under the app/object_detection/configs dir.
    """
    def __init__(self, config=None):
        self.config = config

    def data_preprocess(self, input):
        r"""
        Preprocess the input image.

        Args:
            input (numpy.ndarray): the input image.

        Returns:
            list, the preprocess image shape.
            numpy.ndarray, the preprocess image result.
        """
        if not isinstance(input, np.ndarray):
            err_msg = 'The input type should be numpy.ndarray, got {}.'.format(type(input))
            raise TypeError(err_msg)
        image_height, image_width, _ = input.shape
        image_shape = (image_height, image_width)
        cvt_input = cv2.cvtColor(input, cv2.COLOR_BGR2RGB)

        transform_func = transform_checker.get(self.config.get('dataset'))
        if not transform_func:
            err_msg = 'Currently dataset only supports {} transform!'.format(str(list(transform_checker.keys())))
            raise KeyError(err_msg)
        transform_input = transform_func(cvt_input)
        return image_shape, transform_input

    def convert2tensor(self, transform_input):
        r"""
        Convert the numpy data to the tensor format.

        Args:
            transform_input (numpy.ndarray): the preprocessing image.

        Returns:
            Tensor, the converted image.
        """
        if not isinstance(transform_input, np.ndarray):
            err_msg = 'The transform_input type should be numpy.ndarray, got {}.'.format(type(transform_input))
            raise TypeError(err_msg)
        input_tensor = ts.expand_dims(ts.array(list(transform_input)), 0)
        return input_tensor

    def model_build(self, is_training=False):
        r"""
        Build the object detection model to predict the image.

        Args:
            is_training (bool): default: False.

        Returns:
            model.Model, generated object detection model.
        """
        model_net = model_checker.get(self.config.get('model_net'))
        if not model_net:
            err_msg = 'Currently model_net only supports {}!'.format(str(list(model_checker.keys())))
            raise KeyError(err_msg)

        class_num = self.config.get('class_num')
        if class_num <= 0:
            err_msg = 'The class_num should be an integer greater than 0, got {}.'.format(class_num)
            raise ValueError(err_msg)

        net = model_net(class_num=class_num, is_training=is_training)
        serve_model = model.Model(net)
        return serve_model

    def model_load_and_predict(self, serve_model, input_tensor):
        r"""
        Load the object detection model to predict the image.

        Args:
            serve_model (model.Model): object detection model.
            input_tensor (Tensor): the converted input image.

        Returns:
            model.Model, object detection model loaded the checkpoint file.
            list, predictions output result.
        """
        ckpt_path = self.config.get('checkpoint_path')
        if not ckpt_path:
            err_msg = 'The ckpt_path {} can not be none.'.format(ckpt_path)
            raise TypeError(err_msg)

        ckpt_name = self.config.get('checkpoint_name')
        if not ckpt_name.endswith('.ckpt'):
            err_msg = 'Currently model only supports `ckpt` format, got {}.'.format(ckpt_name)
            raise TypeError(err_msg)

        ckpt_file = os.path.join(ckpt_path, ckpt_name)
        if not os.path.isfile(ckpt_file):
            raise FileNotFoundError("The model checkpoint file path {} does not exist!".format(ckpt_file))
        serve_model.load_checkpoint(ckpt_file)

        predictions_output = serve_model.predict(input_tensor)
        return serve_model, predictions_output

    def data_postprocess(self, predictions_output, image_shape):
        r"""
        Postprocessing the predictions output data.

        Args:
            predictions_output (list): predictions output data.
            image_shape (list): the shape of the input image.

        Returns:
            dict, the postprocessing result.
        """
        output_np = (ts.concatenate((predictions_output[0], predictions_output[1]), axis=-1).asnumpy())
        transform_func = transform_checker.get(self.config.get('dataset'))
        if not transform_func:
            raise KeyError("Currently dataset only supports {} transform!".format(str(list(transform_checker.keys()))))
        bbox_data = transform_func.postprocess(output_np, image_shape)
        return bbox_data


def object_detection_predict(input, object_detector, is_training=False):
    r"""
    An easy object detection model predicting method for beginning developers to use.

    Args:
        input (numpy.ndarray): the input image.
        object_detector (ObjectDetector): the instance of the ObjectDetector class.
        is_training (bool): default: False.

    Returns:
        dict, the postprocessing result.
    """
    if not isinstance(object_detector, ObjectDetector):
        err_msg = 'The object_detector is not the instance of ObjectDetector'
        raise TypeError(err_msg)
    image_shape, transform_input = object_detector.data_preprocess(input)
    input_tensor = object_detector.convert2tensor(transform_input)
    serve_model = object_detector.model_build(is_training=is_training)
    _, predictions_output = object_detector.model_load_and_predict(serve_model, input_tensor)
    detection_bbox_data = object_detector.data_postprocess(predictions_output, image_shape)
    return detection_bbox_data



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import os
import numpy as np
import tinyms as ts
from mindspore.train.serialization import load_checkpoint, load_param_into_net
from tinyms import Tensor


def cyclegan_predict(G_generator, input_data, ckpt_path):
    """
    Single image predict.

    Args:
       G_generator (Generator): Generator, such as 'G_A'.
       input_data (Tensor): input image data.
       ckpt_path (str): the checkpoint path.

    Returns:
        Tensor, generated fake image data.

    Raises:
        ValueError: If `ckpt_path` does not exist.
        ValueError: If `fake_img` is not Tensor or Numpy.
    """
    G_generator.set_train(True)
    # load checkpoint
    if not os.path.isfile(ckpt_path):
        err_msg = "The model path " + ckpt_path + " not exist!"
        raise ValueError(err_msg)
    param_G = load_checkpoint(ckpt_path)
    load_param_into_net(G_generator, param_G)
    fake_img = G_generator(ts.expand_dims(input_data, 0))
    if isinstance(fake_img, Tensor):
        # Decode a [1, C, H, W] Tensor to image numpy array.
        mean = 0.5 * 255
        std = 0.5 * 255
        fake_img = (fake_img.asnumpy()[0] * std + mean).astype(np.uint8).transpose((1, 2, 0))
    elif not isinstance(fake_img, np.ndarray):
        raise ValueError("img should be Tensor or numpy array, but get {}".format(type(fake_img)))
    return fake_img



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""DeepFM Init."""

from .deepfm_base_model import deepfm, DeepFM
from .deepfm_eval_model import DeepFMEvalModel
from .deepfm_train_model import DeepFMWithLoss, DeepFMTrainModel


__all__ = [
    'deepfm', 'DeepFM',
    'DeepFMEvalModel',
    'DeepFMWithLoss', 'DeepFMTrainModel'
]



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Bert model."""

import math
import copy
import numpy as np

import tinyms as ts
from tinyms.initializers import TruncatedNormal, initializer
from tinyms import layers
from tinyms import primitives as P
from tinyms import Tensor
from tinyms import Parameter


class EmbeddingLookup(layers.Layer):
    """
    A embeddings lookup table with a fixed dictionary and size.

    Args:
        vocab_size (int): Size of the dictionary of embeddings.
        embedding_size (int): The size of each embedding vector.
        embedding_shape (list): [batch_size, seq_length, embedding_size], the shape of
                         each embedding vector.
        use_one_hot_embeddings (bool): Specifies whether to use one hot encoding form. Default: False.
        initializer_range (float): Initialization value of TruncatedNormal. Default: 0.02.
    """
    def __init__(self,
                 vocab_size,
                 embedding_size,
                 embedding_shape,
                 use_one_hot_embeddings=False,
                 initializer_range=0.02):
        super(EmbeddingLookup, self).__init__()
        self.vocab_size = vocab_size
        self.use_one_hot_embeddings = use_one_hot_embeddings
        self.embedding_table = Parameter(initializer
                                         (TruncatedNormal(initializer_range),
                                          [vocab_size, embedding_size]))
        self.expand = P.ExpandDims()
        self.shape_flat = (-1,)
        self.gather = P.Gather()
        self.one_hot = P.OneHot()
        self.on_value = Tensor(1.0, ts.float32)
        self.off_value = Tensor(0.0, ts.float32)
        self.array_mul = P.MatMul()
        self.reshape = P.Reshape()
        self.shape = tuple(embedding_shape)

    def construct(self, input_ids):
        """Get output and embeddings lookup table"""
        extended_ids = self.expand(input_ids, -1)
        flat_ids = self.reshape(extended_ids, self.shape_flat)
        if self.use_one_hot_embeddings:
            one_hot_ids = self.one_hot(flat_ids, self.vocab_size, self.on_value, self.off_value)
            output_for_reshape = self.array_mul(
                one_hot_ids, self.embedding_table)
        else:
            output_for_reshape = self.gather(self.embedding_table, flat_ids, 0)
        output = self.reshape(output_for_reshape, self.shape)
        return output, self.embedding_table


class EmbeddingPostprocessor(layers.Layer):
    """
    Postprocessors apply positional and token type embeddings to word embeddings.

    Args:
        embedding_size (int): The size of each embedding vector.
        embedding_shape (list): [batch_size, seq_length, embedding_size], the shape of
                         each embedding vector.
        use_token_type (bool): Specifies whether to use token type embeddings. Default: False.
        token_type_vocab_size (int): Size of token type vocab. Default: 16.
        use_one_hot_embeddings (bool): Specifies whether to use one hot encoding form. Default: False.
        initializer_range (float): Initialization value of TruncatedNormal. Default: 0.02.
        max_position_embeddings (int): Maximum length of sequences used in this
                                 model. Default: 512.
        dropout_prob (float): The dropout probability. Default: 0.1.
    """
    def __init__(self,
                 embedding_size,
                 embedding_shape,
                 use_relative_positions=False,
                 use_token_type=False,
                 token_type_vocab_size=16,
                 use_one_hot_embeddings=False,
                 initializer_range=0.02,
                 max_position_embeddings=512,
                 dropout_prob=0.1):
        super(EmbeddingPostprocessor, self).__init__()
        self.use_token_type = use_token_type
        self.token_type_vocab_size = token_type_vocab_size
        self.use_one_hot_embeddings = use_one_hot_embeddings
        self.max_position_embeddings = max_position_embeddings
        self.embedding_table = Parameter(initializer
                                         (TruncatedNormal(initializer_range),
                                          [token_type_vocab_size,
                                           embedding_size]),
                                         name='embedding_table')

        self.shape_flat = (-1,)
        self.one_hot = layers.OneHot()
        self.on_value = Tensor(1.0, ts.float32)
        self.off_value = Tensor(0.1, ts.float32)
        self.array_mul = P.MatMul()
        self.reshape = P.Reshape()
        self.shape = tuple(embedding_shape)
        self.layernorm = layers.LayerNorm((embedding_size,))
        self.dropout = layers.Dropout(1 - dropout_prob)
        self.gather = P.Gather()
        self.use_relative_positions = use_relative_positions
        self.slice = P.StridedSlice()
        self.full_position_embeddings = Parameter(initializer
                                                  (TruncatedNormal(initializer_range),
                                                   [max_position_embeddings,
                                                    embedding_size]),
                                                  name='full_position_embeddings')

    def construct(self, token_type_ids, word_embeddings):
        """Postprocessors apply positional and token type embeddings to word embeddings."""
        output = word_embeddings
        if self.use_token_type:
            flat_ids = self.reshape(token_type_ids, self.shape_flat)
            if self.use_one_hot_embeddings:
                one_hot_ids = self.one_hot(flat_ids,
                                           self.token_type_vocab_size, self.on_value, self.off_value)
                token_type_embeddings = self.array_mul(one_hot_ids,
                                                       self.embedding_table)
            else:
                token_type_embeddings = self.gather(self.embedding_table, flat_ids, 0)
            token_type_embeddings = self.reshape(token_type_embeddings, self.shape)
            output += token_type_embeddings
        if not self.use_relative_positions:
            _, seq, width = self.shape
            position_embeddings = self.slice(self.full_position_embeddings, (0, 0), (seq, width), (1, 1))
            position_embeddings = self.reshape(position_embeddings, (1, seq, width))
            output += position_embeddings
        output = self.layernorm(output)
        output = self.dropout(output)
        return output


class BertOutput(layers.Layer):
    """
    Apply a linear computation to hidden status and a residual computation to input.

    Args:
        in_channels (int): Input channels.
        out_channels (int): Output channels.
        initializer_range (float): Initialization value of TruncatedNormal. Default: 0.02.
        dropout_prob (float): The dropout probability. Default: 0.1.
        compute_type (:class:`tinyms.dtype`): Compute type in BertTransformer. Default: ts.float32.
    """
    def __init__(self,
                 in_channels,
                 out_channels,
                 initializer_range=0.02,
                 dropout_prob=0.1,
                 compute_type=ts.float32):
        super(BertOutput, self).__init__()
        self.dense = layers.Dense(in_channels, out_channels,
                              weight_init=TruncatedNormal(initializer_range)).to_float(compute_type)
        self.dropout = layers.Dropout(1 - dropout_prob)
        self.dropout_prob = dropout_prob
        self.add = P.Add()
        self.layernorm = layers.LayerNorm((out_channels,)).to_float(compute_type)
        self.cast = P.Cast()


    def construct(self, hidden_status, input_tensor):
        output = self.dense(hidden_status)
        output = self.dropout(output)
        output = self.add(input_tensor, output)
        output = self.layernorm(output)
        return output


class RelaPosMatrixGenerator(layers.Layer):
    """
    Generates matrix of relative positions between inputs.

    Args:
        length (int): Length of one dim for the matrix to be generated.
        max_relative_position (int): Max value of relative position.
    """
    def __init__(self, length, max_relative_position):
        super(RelaPosMatrixGenerator, self).__init__()
        self._length = length
        self._max_relative_position = max_relative_position
        self._min_relative_position = -max_relative_position
        self.range_length = -length + 1

        self.tile = P.Tile()
        self.range_mat = P.Reshape()
        self.sub = P.Sub()
        self.expanddims = P.ExpandDims()
        self.cast = P.Cast()

    def construct(self):
        """Generates matrix of relative positions between inputs."""
        range_vec_row_out = self.cast(ts.array(ts.arange(self._length)), int32)
        range_vec_col_out = self.range_mat(range_vec_row_out, (self._length, -1))
        tile_row_out = self.tile(range_vec_row_out, (self._length,))
        tile_col_out = self.tile(range_vec_col_out, (1, self._length))
        range_mat_out = self.range_mat(tile_row_out, (self._length, self._length))
        transpose_out = self.range_mat(tile_col_out, (self._length, self._length))
        distance_mat = self.sub(range_mat_out, transpose_out)

        distance_mat_clipped = P.clip_by_value(distance_mat,
                                               self._min_relative_position,
                                               self._max_relative_position)

        # Shift values to be >=0. Each integer still uniquely identifies a
        # relative position difference.
        final_mat = distance_mat_clipped + self._max_relative_position
        return final_mat


class RelaPosEmbeddingsGenerator(layers.Layer):
    """
    Generates tensor of size [length, length, depth].

    Args:
        length (int): Length of one dim for the matrix to be generated.
        depth (int): Size of each attention head.
        max_relative_position (int): Maxmum value of relative position.
        initializer_range (float): Initialization value of TruncatedNormal.
        use_one_hot_embeddings (bool): Specifies whether to use one hot encoding form. Default: False.
    """
    def __init__(self,
                 length,
                 depth,
                 max_relative_position,
                 initializer_range,
                 use_one_hot_embeddings=False):
        super(RelaPosEmbeddingsGenerator, self).__init__()
        self.depth = depth
        self.vocab_size = max_relative_position * 2 + 1
        self.use_one_hot_embeddings = use_one_hot_embeddings

        self.embeddings_table = Parameter(
            initializer(TruncatedNormal(initializer_range),
                        [self.vocab_size, self.depth]))

        self.relative_positions_matrix = RelaPosMatrixGenerator(length=length,
                                                                max_relative_position=max_relative_position)
        self.reshape = P.Reshape()
        self.one_hot = layers.OneHot(depth=self.vocab_size)
        self.shape = P.Shape()
        self.gather = P.Gather()  # index_select
        self.matmul = P.BatchMatMul()

    def construct(self):
        """Generate embedding for each relative position of dimension depth."""
        relative_positions_matrix_out = self.relative_positions_matrix()

        if self.use_one_hot_embeddings:
            flat_relative_positions_matrix = self.reshape(relative_positions_matrix_out, (-1,))
            one_hot_relative_positions_matrix = self.one_hot(
                flat_relative_positions_matrix)
            embeddings = self.matmul(one_hot_relative_positions_matrix, self.embeddings_table)
            my_shape = self.shape(relative_positions_matrix_out) + (self.depth,)
            embeddings = self.reshape(embeddings, my_shape)
        else:
            embeddings = self.gather(self.embeddings_table,
                                     relative_positions_matrix_out, 0)
        return embeddings


class SaturateCast(layers.Layer):
    """
    Performs a safe saturating cast. This operation applies proper clamping before casting to prevent
    the danger that the value will overflow or underflow.

    Args:
        src_type (:class:`tinyms.dtype`): The type of the elements of the input tensor. Default: ts.float32.
        dst_type (:class:`tinyms.dtype`): The type of the elements of the output tensor. Default: ts.float32.
    """
    def __init__(self, src_type=ts.float32, dst_type=ts.float32):
        super(SaturateCast, self).__init__()
        np_type = ts.dtype_to_nptype(dst_type)

        self.tensor_min_type = float(np.finfo(np_type).min)
        self.tensor_max_type = float(np.finfo(np_type).max)

        self.min_op = P.Minimum()
        self.max_op = P.Maximum()
        self.cast = P.Cast()
        self.dst_type = dst_type

    def construct(self, x):
        out = self.max_op(x, self.tensor_min_type)
        out = self.min_op(out, self.tensor_max_type)
        return self.cast(out, self.dst_type)


class BertAttention(layers.Layer):
    """
    Apply multi-headed attention from "from_tensor" to "to_tensor".

    Args:
        from_tensor_width (int): Size of last dim of from_tensor.
        to_tensor_width (int): Size of last dim of to_tensor.
        from_seq_length (int): Length of from_tensor sequence.
        to_seq_length (int): Length of to_tensor sequence.
        num_attention_heads (int): Number of attention heads. Default: 1.
        size_per_head (int): Size of each attention head. Default: 512.
        query_act (str): Activation function for the query transform. Default: None.
        key_act (str): Activation function for the key transform. Default: None.
        value_act (str): Activation function for the value transform. Default: None.
        has_attention_mask (bool): Specifies whether to use attention mask. Default: False.
        attention_probs_dropout_prob (float): The dropout probability for
                                      BertAttention. Default: 0.0.
        use_one_hot_embeddings (bool): Specifies whether to use one hot encoding form. Default: False.
        initializer_range (float): Initialization value of TruncatedNormal. Default: 0.02.
        do_return_2d_tensor (bool): True for return 2d tensor. False for return 3d
                             tensor. Default: False.
        use_relative_positions (bool): Specifies whether to use relative positions. Default: False.
        compute_type (:class:`tinyms.dtype`): Compute type in BertAttention. Default: ts.float32.
    """
    def __init__(self,
                 from_tensor_width,
                 to_tensor_width,
                 from_seq_length,
                 to_seq_length,
                 num_attention_heads=1,
                 size_per_head=512,
                 query_act=None,
                 key_act=None,
                 value_act=None,
                 has_attention_mask=False,
                 attention_probs_dropout_prob=0.0,
                 use_one_hot_embeddings=False,
                 initializer_range=0.02,
                 do_return_2d_tensor=False,
                 use_relative_positions=False,
                 compute_type=ts.float32):

        super(BertAttention, self).__init__()
        self.from_seq_length = from_seq_length
        self.to_seq_length = to_seq_length
        self.num_attention_heads = num_attention_heads
        self.size_per_head = size_per_head
        self.has_attention_mask = has_attention_mask
        self.use_relative_positions = use_relative_positions

        self.scores_mul = 1.0 / math.sqrt(float(self.size_per_head))
        self.reshape = P.Reshape()
        self.shape_from_2d = (-1, from_tensor_width)
        self.shape_to_2d = (-1, to_tensor_width)
        weight = TruncatedNormal(initializer_range)
        units = num_attention_heads * size_per_head
        self.query_layer = layers.Dense(from_tensor_width,
                                    units,
                                    activation=query_act,
                                    weight_init=weight).to_float(compute_type)
        self.key_layer = layers.Dense(to_tensor_width,
                                  units,
                                  activation=key_act,
                                  weight_init=weight).to_float(compute_type)
        self.value_layer = layers.Dense(to_tensor_width,
                                    units,
                                    activation=value_act,
                                    weight_init=weight).to_float(compute_type)

        self.shape_from = (-1, from_seq_length, num_attention_heads, size_per_head)
        self.shape_to = (-1, to_seq_length, num_attention_heads, size_per_head)

        self.matmul_trans_b = P.BatchMatMul(transpose_b=True)
        self.multiply = P.Mul()
        self.transpose = P.Transpose()
        self.trans_shape = (0, 2, 1, 3)
        self.trans_shape_relative = (2, 0, 1, 3)
        self.trans_shape_position = (1, 2, 0, 3)
        self.multiply_data = -10000.0
        self.matmul = P.BatchMatMul()

        self.softmax = layers.Softmax()
        self.dropout = layers.Dropout(1 - attention_probs_dropout_prob)

        if self.has_attention_mask:
            self.expand_dims = P.ExpandDims()
            self.sub = P.Sub()
            self.add = P.Add()
            self.cast = P.Cast()
            self.get_dtype = P.DType()
        if do_return_2d_tensor:
            self.shape_return = (-1, num_attention_heads * size_per_head)
        else:
            self.shape_return = (-1, from_seq_length, num_attention_heads * size_per_head)

        self.cast_compute_type = SaturateCast(dst_type=compute_type)
        if self.use_relative_positions:
            self._generate_relative_positions_embeddings = \
                RelaPosEmbeddingsGenerator(length=to_seq_length,
                                           depth=size_per_head,
                                           max_relative_position=16,
                                           initializer_range=initializer_range,
                                           use_one_hot_embeddings=use_one_hot_embeddings)

    def construct(self, from_tensor, to_tensor, attention_mask):
        """reshape 2d/3d input tensors to 2d"""
        from_tensor_2d = self.reshape(from_tensor, self.shape_from_2d)
        to_tensor_2d = self.reshape(to_tensor, self.shape_to_2d)
        query_out = self.query_layer(from_tensor_2d)
        key_out = self.key_layer(to_tensor_2d)
        value_out = self.value_layer(to_tensor_2d)

        query_layer = self.reshape(query_out, self.shape_from)
        query_layer = self.transpose(query_layer, self.trans_shape)
        key_layer = self.reshape(key_out, self.shape_to)
        key_layer = self.transpose(key_layer, self.trans_shape)

        attention_scores = self.matmul_trans_b(query_layer, key_layer)

        # use_relative_position, supplementary logic
        if self.use_relative_positions:
            # relations_keys is [F|T, F|T, H]
            relations_keys = self._generate_relative_positions_embeddings()
            relations_keys = self.cast_compute_type(relations_keys)
            # query_layer_t is [F, B, N, H]
            query_layer_t = self.transpose(query_layer, self.trans_shape_relative)
            # query_layer_r is [F, B * N, H]
            query_layer_r = self.reshape(query_layer_t,
                                         (self.from_seq_length,
                                          -1,
                                          self.size_per_head))
            # key_position_scores is [F, B * N, F|T]
            key_position_scores = self.matmul_trans_b(query_layer_r,
                                                      relations_keys)
            # key_position_scores_r is [F, B, N, F|T]
            key_position_scores_r = self.reshape(key_position_scores,
                                                 (self.from_seq_length,
                                                  -1,
                                                  self.num_attention_heads,
                                                  self.from_seq_length))
            # key_position_scores_r_t is [B, N, F, F|T]
            key_position_scores_r_t = self.transpose(key_position_scores_r,
                                                     self.trans_shape_position)
            attention_scores = attention_scores + key_position_scores_r_t

        attention_scores = self.multiply(self.scores_mul, attention_scores)

        if self.has_attention_mask:
            attention_mask = self.expand_dims(attention_mask, 1)
            multiply_out = self.sub(self.cast(ts.array((1.0,)), self.get_dtype(attention_scores)),
                                    self.cast(attention_mask, self.get_dtype(attention_scores)))

            adder = self.multiply(multiply_out, self.multiply_data)
            attention_scores = self.add(adder, attention_scores)

        attention_probs = self.softmax(attention_scores)
        attention_probs = self.dropout(attention_probs)

        value_layer = self.reshape(value_out, self.shape_to)
        value_layer = self.transpose(value_layer, self.trans_shape)
        context_layer = self.matmul(attention_probs, value_layer)

        # use_relative_position, supplementary logic
        if self.use_relative_positions:
            # relations_values is [F|T, F|T, H]
            relations_values = self._generate_relative_positions_embeddings()
            relations_values = self.cast_compute_type(relations_values)
            # attention_probs_t is [F, B, N, T]
            attention_probs_t = self.transpose(attention_probs, self.trans_shape_relative)
            # attention_probs_r is [F, B * N, T]
            attention_probs_r = self.reshape(
                attention_probs_t,
                (self.from_seq_length,
                 -1,
                 self.to_seq_length))
            # value_position_scores is [F, B * N, H]
            value_position_scores = self.matmul(attention_probs_r,
                                                relations_values)
            # value_position_scores_r is [F, B, N, H]
            value_position_scores_r = self.reshape(value_position_scores,
                                                   (self.from_seq_length,
                                                    -1,
                                                    self.num_attention_heads,
                                                    self.size_per_head))
            # value_position_scores_r_t is [B, N, F, H]
            value_position_scores_r_t = self.transpose(value_position_scores_r,
                                                       self.trans_shape_position)
            context_layer = context_layer + value_position_scores_r_t

        context_layer = self.transpose(context_layer, self.trans_shape)
        context_layer = self.reshape(context_layer, self.shape_return)

        return context_layer


class BertSelfAttention(layers.Layer):
    """
    Apply self-attention.

    Args:
        seq_length (int): Length of input sequence.
        hidden_size (int): Size of the bert encoder layers.
        num_attention_heads (int): Number of attention heads. Default: 12.
        attention_probs_dropout_prob (float): The dropout probability for
                                      BertAttention. Default: 0.1.
        use_one_hot_embeddings (bool): Specifies whether to use one_hot encoding form. Default: False.
        initializer_range (float): Initialization value of TruncatedNormal. Default: 0.02.
        hidden_dropout_prob (float): The dropout probability for BertOutput. Default: 0.1.
        use_relative_positions (bool): Specifies whether to use relative positions. Default: False.
        compute_type (:class:`tinyms.dtype`): Compute type in BertSelfAttention. Default: ts.float32.
    """
    def __init__(self,
                 seq_length,
                 hidden_size,
                 num_attention_heads=12,
                 attention_probs_dropout_prob=0.1,
                 use_one_hot_embeddings=False,
                 initializer_range=0.02,
                 hidden_dropout_prob=0.1,
                 use_relative_positions=False,
                 compute_type=ts.float32):
        super(BertSelfAttention, self).__init__()
        if hidden_size % num_attention_heads != 0:
            raise ValueError("The hidden size (%d) is not a multiple of the number "
                             "of attention heads (%d)" % (hidden_size, num_attention_heads))

        self.size_per_head = int(hidden_size / num_attention_heads)

        self.attention = BertAttention(
            from_tensor_width=hidden_size,
            to_tensor_width=hidden_size,
            from_seq_length=seq_length,
            to_seq_length=seq_length,
            num_attention_heads=num_attention_heads,
            size_per_head=self.size_per_head,
            attention_probs_dropout_prob=attention_probs_dropout_prob,
            use_one_hot_embeddings=use_one_hot_embeddings,
            initializer_range=initializer_range,
            use_relative_positions=use_relative_positions,
            has_attention_mask=True,
            do_return_2d_tensor=True,
            compute_type=compute_type)

        self.output = BertOutput(in_channels=hidden_size,
                                 out_channels=hidden_size,
                                 initializer_range=initializer_range,
                                 dropout_prob=hidden_dropout_prob,
                                 compute_type=compute_type)
        self.reshape = P.Reshape()
        self.shape = (-1, hidden_size)

    def construct(self, input_tensor, attention_mask):
        input_tensor = self.reshape(input_tensor, self.shape)
        attention_output = self.attention(input_tensor, input_tensor, attention_mask)
        output = self.output(attention_output, input_tensor)
        return output


class BertEncoderLayer(layers.Layer):
    """
    Encoder Layer used in BertTransformer.

    Args:
        hidden_size (int): Size of the bert encoder layers. Default: 768.
        seq_length (int): Length of input sequence. Default: 512.
        num_attention_heads (int): Number of attention heads. Default: 12.
        intermediate_size (int): Size of intermediate layer. Default: 3072.
        attention_probs_dropout_prob (float): The dropout probability for
                                      BertAttention. Default: 0.02.
        use_one_hot_embeddings (bool): Specifies whether to use one hot encoding form. Default: False.
        initializer_range (float): Initialization value of TruncatedNormal. Default: 0.02.
        hidden_dropout_prob (float): The dropout probability for BertOutput. Default: 0.1.
        use_relative_positions (bool): Specifies whether to use relative positions. Default: False.
        hidden_act (str): Activation function. Default: "gelu".
        compute_type (:class:`tinyms.dtype`): Compute type in attention. Default: ts.float32.
    """
    def __init__(self,
                 hidden_size=768,
                 seq_length=512,
                 num_attention_heads=12,
                 intermediate_size=3072,
                 attention_probs_dropout_prob=0.02,
                 use_one_hot_embeddings=False,
                 initializer_range=0.02,
                 hidden_dropout_prob=0.1,
                 use_relative_positions=False,
                 hidden_act="gelu",
                 compute_type=ts.float32):
        super(BertEncoderLayer, self).__init__()
        self.attention = BertSelfAttention(
            hidden_size=hidden_size,
            seq_length=seq_length,
            num_attention_heads=num_attention_heads,
            attention_probs_dropout_prob=attention_probs_dropout_prob,
            use_one_hot_embeddings=use_one_hot_embeddings,
            initializer_range=initializer_range,
            hidden_dropout_prob=hidden_dropout_prob,
            use_relative_positions=use_relative_positions,
            compute_type=compute_type)
        self.intermediate = layers.Dense(in_channels=hidden_size,
                                     out_channels=intermediate_size,
                                     activation=hidden_act,
                                     weight_init=TruncatedNormal(initializer_range)).to_float(compute_type)
        self.output = BertOutput(in_channels=intermediate_size,
                                 out_channels=hidden_size,
                                 initializer_range=initializer_range,
                                 dropout_prob=hidden_dropout_prob,
                                 compute_type=compute_type)

    def construct(self, hidden_states, attention_mask):
        # self-attention
        attention_output = self.attention(hidden_states, attention_mask)
        # feed construct
        intermediate_output = self.intermediate(attention_output)
        # add and normalize
        output = self.output(intermediate_output, attention_output)
        return output


class BertTransformer(layers.Layer):
    """
    Multi-layer bert transformer.

    Args:
        hidden_size (int): Size of the encoder layers.
        seq_length (int): Length of input sequence.
        num_hidden_layers (int): Number of hidden layers in encoder layers.
        num_attention_heads (int): Number of attention heads in encoder layers. Default: 12.
        intermediate_size (int): Size of intermediate layer in encoder layers. Default: 3072.
        attention_probs_dropout_prob (float): The dropout probability for
                                      BertAttention. Default: 0.1.
        use_one_hot_embeddings (bool): Specifies whether to use one hot encoding form. Default: False.
        initializer_range (float): Initialization value of TruncatedNormal. Default: 0.02.
        hidden_dropout_prob (float): The dropout probability for BertOutput. Default: 0.1.
        use_relative_positions (bool): Specifies whether to use relative positions. Default: False.
        hidden_act (str): Activation function used in the encoder layers. Default: "gelu".
        compute_type (:class:`tinyms.dtype`): Compute type in BertTransformer. Default: ts.float32.
        return_all_encoders (bool): Specifies whether to return all encoders. Default: False.
    """
    def __init__(self,
                 hidden_size,
                 seq_length,
                 num_hidden_layers,
                 num_attention_heads=12,
                 intermediate_size=3072,
                 attention_probs_dropout_prob=0.1,
                 use_one_hot_embeddings=False,
                 initializer_range=0.02,
                 hidden_dropout_prob=0.1,
                 use_relative_positions=False,
                 hidden_act="gelu",
                 compute_type=ts.float32,
                 return_all_encoders=False):
        super(BertTransformer, self).__init__()
        self.return_all_encoders = return_all_encoders

        slayers = []
        for _ in range(num_hidden_layers):
            layer = BertEncoderLayer(hidden_size=hidden_size,
                                    seq_length=seq_length,
                                    num_attention_heads=num_attention_heads,
                                    intermediate_size=intermediate_size,
                                    attention_probs_dropout_prob=attention_probs_dropout_prob,
                                    use_one_hot_embeddings=use_one_hot_embeddings,
                                    initializer_range=initializer_range,
                                    hidden_dropout_prob=hidden_dropout_prob,
                                    use_relative_positions=use_relative_positions,
                                    hidden_act=hidden_act,
                                    compute_type=compute_type)
            slayers.append(layer)

        self.layers = layers.LayerList(slayers)

        self.reshape = P.Reshape()
        self.shape = (-1, hidden_size)
        self.out_shape = (-1, seq_length, hidden_size)

    def construct(self, input_tensor, attention_mask):
        """Multi-layer bert transformer."""
        prev_output = self.reshape(input_tensor, self.shape)

        all_encoder_layers = ()
        for layer_module in self.layers:
            layer_output = layer_module(prev_output, attention_mask)
            prev_output = layer_output

            if self.return_all_encoders:
                layer_output = self.reshape(layer_output, self.out_shape)
                all_encoder_layers = all_encoder_layers + (layer_output,)

        if not self.return_all_encoders:
            prev_output = self.reshape(prev_output, self.out_shape)
            all_encoder_layers = all_encoder_layers + (prev_output,)
        return all_encoder_layers


class CreateAttentionMaskFromInputMask(layers.Layer):
    """
    Create attention mask according to input mask.

    Args:
        config (Class): Configuration for BertModel.
    """
    def __init__(self, config):
        super(CreateAttentionMaskFromInputMask, self).__init__()
        self.input_mask = None

        self.cast = P.Cast()
        self.reshape = P.Reshape()
        self.shape = (-1, 1, config.seq_length)

    def construct(self, input_mask):
        attention_mask = self.cast(self.reshape(input_mask, self.shape), ts.float32)
        return attention_mask


class Bert(layers.Layer):
    """
    Bidirectional Encoder Representations from Transformers.

    Args:
        config (Class): Configuration for BertModel.
        is_training (bool): True for training mode. False for eval mode.
        use_one_hot_embeddings (bool): Specifies whether to use one hot encoding form. Default: False.
    """
    def __init__(self,
                 config,
                 is_training,
                 use_one_hot_embeddings=False):
        super(Bert, self).__init__()
        config = copy.deepcopy(config)
        if not is_training:
            config.hidden_dropout_prob = 0.0
            config.attention_probs_dropout_prob = 0.0

        self.seq_length = config.seq_length
        self.hidden_size = config.hidden_size
        self.num_hidden_layers = config.num_hidden_layers
        self.embedding_size = config.hidden_size
        self.token_type_ids = None

        self.last_idx = self.num_hidden_layers - 1
        output_embedding_shape = [-1, self.seq_length, self.embedding_size]

        self.bert_embedding_lookup = layers.Embedding(
            vocab_size=config.vocab_size,
            embedding_size=self.embedding_size,
            use_one_hot=use_one_hot_embeddings,
            embedding_table=TruncatedNormal(config.initializer_range))

        self.bert_embedding_postprocessor = EmbeddingPostprocessor(
            embedding_size=self.embedding_size,
            embedding_shape=output_embedding_shape,
            use_relative_positions=config.use_relative_positions,
            use_token_type=True,
            token_type_vocab_size=config.type_vocab_size,
            use_one_hot_embeddings=use_one_hot_embeddings,
            initializer_range=0.02,
            max_position_embeddings=config.max_position_embeddings,
            dropout_prob=config.hidden_dropout_prob)

        self.bert_encoder = BertTransformer(
            hidden_size=self.hidden_size,
            seq_length=self.seq_length,
            num_attention_heads=config.num_attention_heads,
            num_hidden_layers=self.num_hidden_layers,
            intermediate_size=config.intermediate_size,
            attention_probs_dropout_prob=config.attention_probs_dropout_prob,
            use_one_hot_embeddings=use_one_hot_embeddings,
            initializer_range=config.initializer_range,
            hidden_dropout_prob=config.hidden_dropout_prob,
            use_relative_positions=config.use_relative_positions,
            hidden_act=config.hidden_act,
            compute_type=config.compute_type,
            return_all_encoders=True)

        self.cast = P.Cast()
        self.dtype = config.dtype
        self.cast_compute_type = SaturateCast(dst_type=config.compute_type)
        self.slice = P.StridedSlice()

        self.squeeze_1 = P.Squeeze(axis=1)
        self.dense = layers.Dense(self.hidden_size, self.hidden_size,
                              activation="tanh",
                              weight_init=TruncatedNormal(config.initializer_range)).to_float(config.compute_type)
        self._create_attention_mask_from_input_mask = CreateAttentionMaskFromInputMask(config)

    def construct(self, input_ids, token_type_ids, input_mask):
        """Bidirectional Encoder Representations from Transformers."""
        # embedding
        embedding_tables = self.bert_embedding_lookup.embedding_table
        word_embeddings = self.bert_embedding_lookup(input_ids)
        embedding_output = self.bert_embedding_postprocessor(token_type_ids,
                                                             word_embeddings)

        # attention mask [batch_size, seq_length, seq_length]
        attention_mask = self._create_attention_mask_from_input_mask(input_mask)

        # bert encoder
        encoder_output = self.bert_encoder(self.cast_compute_type(embedding_output),
                                           attention_mask)

        sequence_output = self.cast(encoder_output[self.last_idx], self.dtype)

        # pooler
        batch_size = P.Shape()(input_ids)[0]
        sequence_slice = self.slice(sequence_output,
                                    (0, 0, 0),
                                    (batch_size, 1, self.hidden_size),
                                    (1, 1, 1))
        first_token = self.squeeze_1(sequence_slice)
        pooled_output = self.dense(first_token)
        pooled_output = self.cast(pooled_output, self.dtype)

        return sequence_output, pooled_output, embedding_tables


def bert(config,is_training,use_one_hot_embeddings=False):
    """
    Get bert neural network.
    """

    return Bert(config=config, is_training=is_training, use_one_hot_embeddings=use_one_hot_embeddings)


# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
Metrics module provides functions to measure the performance of the machine learning
models on the evaluation dataset. It's used to choose the best model.
"""
from mindspore.nn import metrics
from mindspore.nn.metrics import *
from sklearn.metrics import roc_auc_score

__all__ = ['AUCMetric']
__all__.extend(metrics.__all__)


class AUCMetric(Metric):
    """
    Calculates the auc value.
    Implement auc metric method.

    Computes the Area Under the Curve (AUC) using the trapezoidal rule. This is a general function, given points on a
    curve. For computing the area under the ROC-curve.

    Args:
        x (Union[np.array, list]): From the ROC curve(fpr), np.array with false positive rates. If multiclass,
                                   this is a list of such np.array, one for each class. The shape :math:`(N)`.
        y (Union[np.array, list]): From the ROC curve(tpr), np.array with true positive rates. If multiclass,
                                   this is a list of such np.array, one for each class. The shape :math:`(N)`.
        reorder (boolean): If True, assume that the curve is ascending in the case of ties, as for an ROC curve.
                           If the curve is non-ascending, the result will be wrong. Default: False.

    Returns:
        area (float): Compute result.

    Examples:
        >>> from tinyms.metrics import AUCMetric
        >>>
        >>> metric = AUCMetric()
    """
    def __init__(self):
        super(AUCMetric, self).__init__()
        self.pred_probs = []
        self.true_labels = []

    def clear(self):
        """Clear the internal evaluation result."""
        self.pred_probs = []
        self.true_labels = []

    def update(self, *inputs):
        batch_predict = inputs[1].asnumpy()
        batch_label = inputs[2].asnumpy()
        self.pred_probs.extend(batch_predict.flatten().tolist())
        self.true_labels.extend(batch_label.flatten().tolist())

    def eval(self):
        if len(self.true_labels) != len(self.pred_probs):
            raise RuntimeError('true_labels.size() is not equal to pred_probs.size()')
        auc = roc_auc_score(self.true_labels, self.pred_probs)
        return auc



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import numpy as np
from scipy.stats import truncnorm

import tinyms as ts
from tinyms import layers, Tensor
from tinyms.primitives import ReduceMean, Concat
from tinyms.layers import AvgPool2d, ReLU, MaxPool2d


def _conv_variance_scaling_initializer(in_channel, out_channel, kernel_size):
    fan_in = in_channel * kernel_size * kernel_size
    scale = 1.0
    scale /= max(1., fan_in)
    stddev = (scale ** 0.5) / .87962566103423978
    mu, sigma = 0, stddev
    weight = truncnorm(-2, 2, loc=mu, scale=sigma).rvs(out_channel * in_channel * kernel_size * kernel_size)
    return ts.reshape(weight, (out_channel, in_channel, kernel_size, kernel_size))


def _weight_variable(shape, factor=0.01):
    init_value = np.random.randn(*shape).astype(np.float32) * factor
    return Tensor(init_value)


def _conv3x3(in_channel, out_channel, stride=1):
    weight_shape = (out_channel, in_channel, 3, 3)
    weight = _weight_variable(weight_shape)
    return layers.Conv2d(in_channel, out_channel,
                         kernel_size=3, stride=stride, padding=0, pad_mode='same', weight_init=weight)


def _conv1x1(in_channel, out_channel, stride=1):
    weight_shape = (out_channel, in_channel, 1, 1)
    weight = _weight_variable(weight_shape)
    return layers.Conv2d(in_channel, out_channel,
                         kernel_size=1, stride=stride, padding=0, pad_mode='same', weight_init=weight)


def _conv7x7(in_channel, out_channel, stride=1):
    weight_shape = (out_channel, in_channel, 7, 7)
    weight = _weight_variable(weight_shape)
    return layers.Conv2d(in_channel, out_channel,
                         kernel_size=7, stride=stride, padding=0, pad_mode='same', weight_init=weight)


def _bn(channel):
    return layers.BatchNorm2d(channel, eps=1e-4, momentum=0.9,
                              gamma_init=1, beta_init=0, moving_mean_init=0, moving_var_init=1)


def _bn_last(channel):
    return layers.BatchNorm2d(channel, eps=1e-4, momentum=0.9,
                              gamma_init=0, beta_init=0, moving_mean_init=0, moving_var_init=1)


def _fc(in_channel, out_channel):
    weight_shape = (out_channel, in_channel)
    weight = _weight_variable(weight_shape)
    return layers.Dense(in_channel, out_channel, has_bias=True, weight_init=weight, bias_init=0)


class _DenseLayer(layers.Layer):
    def __init__(self, in_channels, growth_rate, bn_size):
        super(_DenseLayer, self).__init__()
        self.layer = layers.SequentialLayer(
            [_bn(in_channels),
             layers.ReLU(),
             _conv1x1(in_channels, bn_size * growth_rate),
             _bn(bn_size*growth_rate),
             layers.ReLU(),
             _conv3x3(bn_size*growth_rate, growth_rate),
             ])
        self.ops = Concat(axis=1)

    def construct(self, x):
        new_features = self.layer(x)
        return self.ops((x, new_features))


class _DenseBlock(layers.Layer):
    def __init__(self, num_layers, in_channels, bn_size, growth_rate):
        super(_DenseBlock, self).__init__()
        self.layer = layers.SequentialLayer()
        for i in range(num_layers):
            self.layer.append(_DenseLayer(in_channels+growth_rate*i,
                                          growth_rate, bn_size))

    def construct(self, x):
        out = self.layer(x)
        return out


class _Transition(layers.Layer):
    def __init__(self, in_channels, out_channels):
        super(_Transition, self).__init__()
        self.layer = layers.SequentialLayer(
            [_bn(in_channels),
             layers.ReLU(),
             _conv1x1(in_channels, out_channels),
             AvgPool2d(kernel_size=2, stride=2, pad_mode='same', data_format='NCHW')
             ])

    def construct(self, x):
        out = self.layer(x)
        return out


class DenseNet(layers.Layer):
    def __init__(self, class_num=1000, growth_rate=12, block_config=(6, 12, 24, 16),
                 bn_size=4, theta=0.5, bc=False):
        super(DenseNet, self).__init__()

        num_init_feature = 2 * growth_rate
        if bc:
            self.features = layers.SequentialLayer(
                [
                    _conv3x3(3, num_init_feature, 1),
                    _bn(num_init_feature),
                    ReLU()
                ]
            )
        else:
            self.features = layers.SequentialLayer(
                [
                    _conv7x7(3, num_init_feature, 2),
                    _bn(num_init_feature),
                    ReLU(),
                    MaxPool2d(kernel_size=2, stride=2, pad_mode='same', data_format='NCHW')
                ]
            )

        num_feature = num_init_feature
        for i, num_layers in enumerate(block_config):

            self.features.append(_DenseBlock(num_layers, num_feature,
                                             bn_size, growth_rate))
            num_feature = num_feature + growth_rate * num_layers
            if i != len(block_config)-1:
                self.features.append(_Transition(num_feature, int(num_feature * theta)))
                num_feature = int(num_feature * theta)

        self.norm = _bn(num_feature)
        self.relu = ReLU()
        # self.features.append([_bn(num_feature),ReLU()])
        self.mean = ReduceMean(keep_dims=True)
        self.flatten = layers.Flatten()
        self.end_point = _fc(num_feature, class_num)

    def construct(self, x):
        features = self.features(x)
        features = self.relu(self.norm(features))
        features = self.end_point(self.flatten(self.mean(features, (2, 3))))
        return features


def densenet100(**kwargs):
    """
    Get DenseNet instance for model training, evaluation and prediction.

    Args:
        class_num (int): The number of classes. Default: 10.

    Returns:
        model.DenseNet, DenseNet instance.
    """
    return DenseNet(class_num=kwargs.get('class_num', 10),
                    growth_rate=12, block_config=(16, 16, 16))



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
This module is to support vision augmentations. transforms is a high performance
image augmentation module which is developed with C++ OpenCV.
"""
from mindspore.dataset.vision.utils import Inter, Border, ImageBatchFormat
from . import transforms
from .transforms import *
from .view import ImageViewer
from .utils import ssd_bboxes_encode, ssd_bboxes_filter, coco_eval
from . import transform_config

vision_utils = ['Inter', 'Border', 'ImageBatchFormat']
bbox_utils = [
    'ssd_bboxes_encode',
    'ssd_bboxes_filter',
    'coco_eval',
]

__all__ = ['ImageViewer']
__all__.extend(vision_utils + bbox_utils)
__all__.extend(transforms.__all__)
__all__.extend(transform_config.__all__)



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

from mindspore import Model as _Model
from mindspore.train.serialization import load_checkpoint as _load_checkpoint, \
    save_checkpoint as _save_checkpoint, export as _export, load as _load

from ..layers import GraphLayer
from .lenet5 import lenet5, LeNet
from .resnet50 import resnet50, ResNet
from .mobilenetv2 import mobilenetv2, MobileNetV2
from .ssd300 import ssd300_mobilenetv2, SSD300
from .cycle_gan.cycle_gan import cycle_gan, cycle_gan_infer
from .densenet100 import densenet100, DenseNet
from .alexnet import alexnet, AlexNet
from .sentimentnet import sentimentnet, SentimentNet
from .bert import bert, Bert
from .vgg import vgg11, vgg13, vgg16, vgg19, VGG
from .deepfm import deepfm, DeepFM, DeepFMEvalModel, DeepFMWithLoss, DeepFMTrainModel

__all__ = [
    'Model',
    'load',
    'lenet5', 'LeNet',
    'resnet50', 'ResNet',
    'mobilenetv2', 'MobileNetV2',
    'ssd300_mobilenetv2', 'SSD300',
    'cycle_gan', 'cycle_gan_infer',
    'densenet100', 'DenseNet',
    'alexnet', 'AlexNet',
    'sentimentnet', 'SentimentNet',
    'bert', 'Bert',
    'vgg11', 'vgg13', 'vgg16', 'vgg19', 'VGG',
    'deepfm', 'DeepFM', 'DeepFMEvalModel', 'DeepFMWithLoss', 'DeepFMTrainModel'
]


class Model(_Model):
    """
    High-Level API for Training or Evaluation.

    `Model` groups layers into an object with training and inference features.

    Args:
        network (layers.Layer): A training or testing network.

    Examples:
        >>> from tinyms.model import Model, lenet5
        >>> form tinyms.losses import SoftmaxCrossEntropyWithLogits
        >>> from tinyms.optimizers import Momentum
        >>>
        >>> net = lenet5(class_num=10)
        >>> model = Model(net)
        >>> net_loss = SoftmaxCrossEntropyWithLogits()
        >>> net_opt = Momentum(params=net.trainable_params(), learning_rate=0.1, momentum=0.9)
        >>> model.compile(loss_fn=net_loss, optimizer=net_opt, metrics=None)
        >>> # For details about how to build the dataset, please refer to the API document on the official website.
        >>> ds_train = create_custom_dataset()
        >>> model.train(2, ds_train)
    """

    def __init__(self, network):
        super(Model, self).__init__(network)

    def compile(self, loss_fn=None, optimizer=None, metrics=None, eval_network=None,
                amp_level="O0", **kwargs):
        """
        High-Level API for configure the train or eval network.

        Args:
            loss_fn (layers.Layer): Objective function, if loss_fn is None, the
                                network should contain the logic of loss and grads calculation, and the logic
                                of parallel if needed. Default: None.
            optimizer (layers.Layer): Optimizer for updating the weights. Default: None.
            metrics (Union[dict, set]): A Dictionary or a set of metrics to be evaluated by the model during
                            training and testing. eg: {'accuracy', 'recall'}. Default: None.
            eval_network (layers.Layer): Network for evaluation. If not defined, `network` and `loss_fn` would be wrapped as
                            `eval_network`. Default: None.
            amp_level (str): Option for argument `level` in `mindspore.amp.build_train_network`, level for mixed
                precision training. Supports ["O0", "O2", "O3", "auto"]. Default: "O0".

                - O0: Do not change.
                - O2: Cast network to float16, keep batchnorm run in float32, using dynamic loss scale.
                - O3: Cast network to float16, with additional property 'keep_batchnorm_fp32=False'.
                - auto: Set to level to recommended level in different devices. Set level to O2 on GPU, Set
                level to O3 Ascend. The recommended level is choose by the export experience, cannot
                always generalize. User should specify the level for special network.

                O2 is recommended on `GPU`, O3 is recommended on `Ascend`.
        """
        # configure the train network
        self._loss_fn = loss_fn
        self._optimizer = optimizer
        self._amp_level = amp_level
        self._check_kwargs(kwargs)
        self._process_amp_args(kwargs)
        self._train_network = self._build_train_network()
        # configure the eval network
        self._build_eval_network(metrics, eval_network=eval_network, eval_indexes=None)
        # configure the predict network
        self._build_predict_network()

    def load_checkpoint(self, ckpt_file_name, strict_load=False):
        """
        Loads checkpoint info from a specified file.

        Args:
            ckpt_file_name (str): Checkpoint file name.
            strict_load (bool): Whether to strict load the parameter into net. If False, it will load parameter
                            in the param_dict into net with the same suffix. Default: False.

        Returns:
            Dict, key is parameter name, value is a Parameter.

        Raises:
            ValueError: Checkpoint file is incorrect.

        Examples:
            >>> ckpt_file_name = "./checkpoint/LeNet5-1_32.ckpt"
            >>> param_dict = model.load_checkpoint(ckpt_file_name)
        """
        return _load_checkpoint(ckpt_file_name, net=self._network,
                                strict_load=strict_load)

    def save_checkpoint(self, ckpt_file_name):
        """
        Saves checkpoint info to a specified file.

        Args:
            ckpt_file_name (str): Checkpoint file name. If the file name already exists, it will be overwritten.

        Raises:
            TypeError: If the parameter save_obj is not layers.Layer or list type.
        """
        return _save_checkpoint(self._network, ckpt_file_name)

    def export(self, inputs, file_name, file_format='MINDIR', **kwargs):
        """
        Export the TinyMS prediction model to a file in the specified format.

        Args:
            inputs (Tensor): Inputs of the `net`.
            file_name (str): File name of the model to be exported.
            file_format (str): MindSpore currently supports `AIR`, `ONNX` and `MINDIR` format for exported model.
                Default: MINDIR.

                - AIR: Ascend Intermediate Representation. An intermediate representation format of Ascend model.
                Recommended suffix for output file is '.air'.

                - ONNX: Open Neural Network eXchange. An open format built to represent machine learning models.
                Recommended suffix for output file is '.onnx'.

                - MINDIR: MindSpore Native Intermediate Representation for Anf. An intermediate representation format
                for MindSpore models.
                Recommended suffix for output file is '.mindir'.

            kwargs (dict): Configuration options dictionary.

                - quant_mode: The mode of quant.
                - mean: Input data mean. Default: 127.5.
                - std_dev: Input data variance. Default: 127.5.
        """
        return _export(self._network, inputs, file_name=file_name,
                       file_format=file_format, **kwargs)


def load(file_name):
    """
    Load MindIR graph and return the network with parameters.

    The returned object is wrapperred by a `GraphLayer`. However, there are some limitations to
    the current use of `GraphLayer`, see class :class:`tinyms.layers.GraphLayer` for more details.

    Args:
        file_name (str): MindIR file name.

    Returns:
        GraphLayer instance, a compiled graph with parameters.

    Raises:
        ValueError: MindIR file is incorrect.

    Examples:
        >>> import tinyms as ts
        >>> import tinyms.layers as layers
        >>> from tinyms.model import Model, load
        >>>
        >>> net = layers.Conv2d(1, 1, kernel_size=3)
        >>> model = Model(net)
        >>> input = ts.ones([1, 1, 3, 3])
        >>> model.export(input, "net", file_format="MINDIR")
        ...
        >>> net = load("net.mindir")
        >>> print(net(input))
        [[[[ 0.02548009  0.04010789  0.03120251]
           [ 0.00268656  0.02353744  0.03807815]
           [-0.00896441 -0.00303641  0.01502199]]]]
    """
    graph = _load(file_name)
    return GraphLayer(graph)



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
.. TinyMS Serving package.

This module refers to the process of serving pre-trained models so that
they can quickly and efficiently process data input by users and obtain
results. TinyMS provides a complete set of start server (`start_server`),
check backend (`list_servables`), check start status (`server_started`)
and shut down the server (`shutdown`) and other functions based on
`Flask` (https://flask.palletsprojects.com/en/1.1.x/).

Examples:
    >>> from tinyms.serving import Server, Client
    >>>
    >>> server = Server()
    >>> server.start_server()
    Server starts at host 127.0.0.1, port 5000
    >>> client = Client()
    >>> client.list_servables()
    >>> client.predict('example.jpg', 'servable_name', dataset_name='mnist')
"""
from . import client, server
from .client import *
from .server import *

__all__ = []
__all__.extend(client.__all__)
__all__.extend(server.__all__)



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
import os
import json

from tinyms.hub.utils.download import download_file_from_url

__all__ = ['load_and parse_config']


def _download_ckeckpoint(checkpoint_url, sha256, checkpoint_path):
    if not checkpoint_url:
        err_msg = 'When set download_from_hub to true, the checkpoint_url can not be empty.'
        raise ValueError(err_msg)

    if not checkpoint_path:
        err_msg = 'When set download_from_hub to true, the checkpoint_path can not be empty.'
        raise ValueError(err_msg)

    if not sha256:
        err_msg = 'When set download_from_hub to true, the sha256 can not be empty.'
        raise ValueError(err_msg)

    download_file_from_url(checkpoint_url, hash_sha256=sha256, save_path=checkpoint_path)


def load_and_parse_config(config_path):
    r"""
    Load and parse the json file the object detection model.

    Args:
        config_path (numpy.ndarray): the config json file path.

    Returns:
        dict, the model configuration.
    """
    # Check if config_path existed
    if not os.path.exists(config_path):
        raise FileNotFoundError("The config file path {} does not exist!".format(config_path))

    with open(config_path, 'r') as f:
        configs = json.load(f)
        if configs.get('download_from_hub'):
            _download_ckeckpoint(configs.get('checkpoint_url'),
                                 configs.get('sha256'),
                                 configs.get('checkpoint_path'))
    return configs



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

import numpy as np

from tinyms import layers, Tensor


def _weight_variable(shape, factor=0.01):
    init_value = np.random.randn(*shape).astype(np.float32) * factor
    return Tensor(init_value)


def _conv3x3(in_channel, out_channel, stride=1):
    weight_shape = (out_channel, in_channel, 3, 3)
    weight = _weight_variable(weight_shape)
    return layers.Conv2d(in_channel, out_channel,
                         kernel_size=3, stride=stride, padding=0, pad_mode='same', weight_init=weight)

class VGG(layers.Layer):
    """
    Get VGG neural network.

    Args:
        features (layers.Layer): Feature extractor.
        class_num (int): Class number. Default: 1000.

    Returns:
        layers.Layer, layer instance of AlexNet neural network.

    Examples:
        >>> from tinyms.model import VGG
        >>>
        >>> net = VGG(features=make_layers(cfg=cfgs['A']),class_num=1000)
    """

    def __init__(self, features, class_num=1000):
        super(VGG, self).__init__()
        self.features = features
        self.flatten = layers.Flatten()
        self.classifier = layers.SequentialLayer([
            layers.Dense(512 * 7 * 7, 4096),
            layers.ReLU(),
            layers.Dropout(),
            layers.Dense(4096, 4096),
            layers.ReLU(),
            layers.Dropout(),
            layers.Dense(4096, class_num),
        ])

    def construct(self, x):
        x = self.features(x)
        x = self.flatten(x)
        x = self.classifier(x)
        return x


def make_layers(cfg, batch_norm=False):
    Layers = []
    in_channels = 3
    for v in cfg:
        if v == 'M':
            Layers += [layers.MaxPool2d(kernel_size=2, stride=2)]
        else:
            conv2d = _conv3x3(in_channels, v)
            if batch_norm:
                Layers += [conv2d, layers.BatchNorm2d(v), layers.ReLU()]
            else:
                Layers += [conv2d, layers.ReLU()]
            in_channels = v
    return layers.SequentialLayer(Layers)


cfgs = {
    'A': [64, 'M', 128, 'M', 256, 256, 'M', 512, 512, 'M', 512, 512, 'M'],
    'B': [64, 64, 'M', 128, 128, 'M', 256, 256, 'M', 512, 512, 'M', 512, 512, 'M'],
    'D': [64, 64, 'M', 128, 128, 'M', 256, 256, 256, 'M', 512, 512, 512, 'M', 512, 512, 512, 'M'],
    'E': [64, 64, 'M', 128, 128, 'M', 256, 256, 256, 256, 'M', 512, 512, 512, 512, 'M', 512, 512, 512, 512, 'M'],
}


def vgg11(**kwargs):
    """
    Get vgg11 neural network.

    Args:
        class_num (int): Class number. Default: 10.
        batch_norm (bool): Whether to use BatchNormalization. Default: True

    Returns:
        layers.Layer, layer instance of vgg11 neural network.

    Examples:
        >>> from tinyms.model import vgg11
        >>>
        >>> net = vgg11(class_num=10)
    """
    return VGG(make_layers(cfg=cfgs['A'], batch_norm=kwargs.get('batch_norm', True)), class_num=kwargs.get('class_num', 10))


def vgg13(**kwargs):
    """
    Get vgg13 neural network.

    Args:
        class_num (int): Class number. Default: 10.
        batch_norm (bool): Whether to use BatchNormalization. Default: True

    Returns:
        layers.Layer, layer instance of vgg13 neural network.

    Examples:
        >>> from tinyms.model import vgg13
        >>>
        >>> net = vgg13(class_num=10)
    """
    return VGG(make_layers(cfg=cfgs['B'], batch_norm=kwargs.get('batch_norm', True)), class_num=kwargs.get('class_num', 10))


def vgg16(**kwargs):
    """
    Get vgg16 neural network.

    Args:
        class_num (int): Class number. Default: 10.
        batch_norm (bool): Whether to use BatchNormalization. Default: True

    Returns:
        layers.Layer, layer instance of vgg16 neural network.

    Examples:
        >>> from tinyms.model import vgg16
        >>>
        >>> net = vgg16(class_num=10)
    """
    return VGG(make_layers(cfg=cfgs['D'], batch_norm=kwargs.get('batch_norm', True)), class_num=kwargs.get('class_num', 10))


def vgg19(**kwargs):
    """
    Get vgg19 neural network.

    Args:
        class_num (int): Class number. Default: 10.
        batch_norm (bool): Whether to use BatchNormalization. Default: True

    Returns:
        layers.Layer, layer instance of vgg19 neural network.

    Examples:
        >>> from tinyms.model import vgg19
        >>>
        >>> net = vgg19(class_num=10)
    """
    return VGG(make_layers(cfg=cfgs['E'], batch_norm=kwargs.get('batch_norm', True)), class_num=kwargs.get('class_num', 10))



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""TinyMS data utils package."""
import os
import sys
import gzip
import tarfile
import requests
import wget
from PIL import Image
import numpy as np
from tinyms import Tensor

__all__ = ['download_dataset', 'generate_image_list', 'load_resized_img', 'load_img']

IMG_EXTENSIONS = ['.jpg', '.jpeg', '.png', '.ppm', '.bmp', '.tif', '.tiff']


def is_image(filename):
    """
    Judge whether it is a picture.

    Args:
        filename (str): image name.

    Returns:
        bool, True or False.

    """
    return any(filename.lower().endswith(extension) for extension in IMG_EXTENSIONS)


def generate_image_list(dir_path, max_dataset_size=float("inf")):
    """
    Traverse the directory to generate a list of images path.

    Args:
        dir_path (str): image directory.
        max_dataset_size (int): Maximum number of return image paths.

    Returns:
        Image path list.

    """
    images = []
    assert os.path.isdir(dir_path), '%s is not a valid directory' % dir_path

    for root, _, fnames in sorted(os.walk(dir_path)):
        for fname in fnames:
            if is_image(fname):
                path = os.path.join(root, fname)
                images.append(path)

    print("len(images):", len(images))
    return images[:min(max_dataset_size, len(images))]


def _unzip(gzip_path):
    """unzip dataset file
    Args:
        gzip_path: dataset file path
    """
    # decompress the file if gzip_path ends with `.tar`
    if gzip_path.endswith('.tar'):
        with tarfile.open(gzip_path) as f:
            f.extractall(gzip_path[:gzip_path.rfind('/')])
    elif gzip_path.endswith('.gz'):
        gzip_file = gzip_path.replace('.gz', '')
        with open(gzip_file, 'wb') as f:
            gz_file = gzip.GzipFile(gzip_path)
            f.write(gz_file.read())
        # decompress the file if gz_file ends with `.tar`
        if gzip_file.endswith('.tar'):
            with tarfile.open(gzip_file) as f:
                f.extractall(gzip_file[:gzip_file.rfind('/')])
    else:
        print("Currently the format of unzip dataset only supports `*.tar`, `*.gz` and `*.tar.gz`!")
        sys.exit(0)


def _fetch_and_unzip(url, file_name):
    """download the dataset from remote url
    Args:
        url: str, remote download url
        file_name: str, local path of downloaded file
    """
    res = requests.get(url, stream=True, verify=False)
    # get dataset size
    total_size = int(res.headers["Content-Length"])
    temp_size = 0
    with open(file_name, "wb+") as f:
        for chunk in res.iter_content(chunk_size=1024):
            temp_size += len(chunk)
            f.write(chunk)
            f.flush()
            done = int(100 * temp_size / total_size)
            # show download progress
            sys.stdout.write("\r[{}{}] {:.2f}%".format("█" * done, " " * (100 - done), 100 * temp_size / total_size))
            sys.stdout.flush()
    print("\n============== {} is ready ==============".format(file_name))
    _unzip(file_name)
    os.remove(file_name)


def _fetch_and_unzip_by_wget(url, file_name):
    """download the dataset from remote url by wget tool
    Args:
        url: str, remote download url
        file_name: str, local path of downloaded file
    """
    # function to show download progress
    def bar_progress(current, total, width=80):
        progress_message = "Downloading: %d%% [%d / %d] bytes" % (current / total * 100, current, total)
        # Don't use print() as it will print in new line every time.
        sys.stdout.write("\r" + progress_message)
        sys.stdout.flush()
    # using wget is faster than fetch.
    wget.download(url, out=file_name, bar=bar_progress)
    print("\n============== {} is ready ==============".format(file_name))
    _unzip(file_name)
    os.remove(file_name)


def _download_mnist(local_path):
    """Download the dataset from http://yann.lecun.com/exdb/mnist/."""
    dataset_path = os.path.join(local_path, 'mnist')
    train_path = os.path.join(dataset_path, 'train')
    test_path = os.path.join(dataset_path, 'test')
    if not os.path.exists(train_path):
        os.makedirs(train_path)
    if not os.path.exists(test_path):
        os.makedirs(test_path)
    print("************** Downloading the MNIST dataset **************")
    train_url = {"http://yann.lecun.com/exdb/mnist/train-images-idx3-ubyte.gz",
                 "http://yann.lecun.com/exdb/mnist/train-labels-idx1-ubyte.gz"}
    test_url = {"http://yann.lecun.com/exdb/mnist/t10k-images-idx3-ubyte.gz",
                "http://yann.lecun.com/exdb/mnist/t10k-labels-idx1-ubyte.gz"}
    for url in train_url:
        # split the file name from url
        file_name = os.path.join(train_path, url.split('/')[-1])
        if not os.path.exists(file_name.replace('.gz', '')):
            _fetch_and_unzip(url, file_name)
    for url in test_url:
        # split the file name from url
        file_name = os.path.join(test_path, url.split('/')[-1])
        if not os.path.exists(file_name.replace('.gz', '')):
            _fetch_and_unzip(url, file_name)

    return dataset_path


def _download_cifar10(local_path):
    '''Download the dataset from http://www.cs.toronto.edu/~kriz/cifar.html.'''
    dataset_path = os.path.join(local_path, 'cifar10')
    if not os.path.exists(dataset_path):
        os.makedirs(dataset_path)

    print("************** Downloading the Cifar10 dataset **************")
    remote_url = "http://www.cs.toronto.edu/~kriz/cifar-10-binary.tar.gz"
    file_name = os.path.join(dataset_path, remote_url.split('/')[-1])
    if not os.path.exists(file_name.replace('.gz', '')):
        _fetch_and_unzip(remote_url, file_name)

    return os.path.join(dataset_path, 'cifar-10-batches-bin')


def _download_cifar100(local_path):
    '''Download the dataset from http://www.cs.toronto.edu/~kriz/cifar.html.'''
    dataset_path = os.path.join(local_path, 'cifar100')
    if not os.path.exists(dataset_path):
        os.makedirs(dataset_path)

    print("************** Downloading the Cifar100 dataset **************")
    remote_url = "http://www.cs.toronto.edu/~kriz/cifar-100-binary.tar.gz"
    file_name = os.path.join(dataset_path, remote_url.split('/')[-1])
    if not os.path.exists(file_name.replace('.gz', '')):
        _fetch_and_unzip(remote_url, file_name)

    return os.path.join(dataset_path, 'cifar-100-binary')


def _download_voc(local_path):
    '''Download the dataset from http://host.robots.ox.ac.uk/pascal/VOC/voc2007/index.html.'''
    dataset_path = os.path.join(local_path, 'voc')
    if not os.path.exists(dataset_path):
        os.makedirs(dataset_path)

    print("************** Downloading the VOC2007 dataset **************")
    remote_url = "http://host.robots.ox.ac.uk/pascal/VOC/voc2007/VOCtrainval_06-Nov-2007.tar"
    file_name = os.path.join(dataset_path, remote_url.split('/')[-1])
    if not os.path.exists(os.path.join(dataset_path, 'VOCdevkit', 'VOC2007')):
        _fetch_and_unzip(remote_url, file_name)

    return os.path.join(dataset_path, 'VOCdevkit', 'VOC2007')


def _check_uncompressed_kaggle_display_advertising_files(dataset_path):
    """check uncompressed kaggle display advertising files."""
    file_name_list = ["train.txt", "test.txt", "readme.txt"]
    file_size_list = [11147184845, 1460246311, 1927]

    for file_name, file_size in zip(file_name_list, file_size_list):
        file_path = os.path.join(dataset_path, file_name)
        if not os.path.exists(file_path):
            return False
        else:
            if os.path.getsize(file_path) != file_size:
                print("************** {} may be error, need to download again **************".
                      format(file_path), flush=True)
                return False

    return True


def _download_kaggle_display_advertising(local_path):
    '''Download the dataset from http://go.criteo.net/criteo-research-kaggle-display-advertising-challenge-dataset.tar.gz.'''
    dataset_path = os.path.join(local_path, "kaggle_display_advertising")
    if not os.path.exists(dataset_path):
        os.makedirs(dataset_path)

    print("************** Downloading the Kaggle Display Advertising Challenge dataset **************")
    remote_url = "http://go.criteo.net/criteo-research-kaggle-display-advertising-challenge-dataset.tar.gz"
    file_name = os.path.join(dataset_path, remote_url.split('/')[-1])
    # already exist criteo-research-kaggle-display-advertising-challenge-dataset.tar.gz
    if os.path.exists(file_name):
        if os.path.getsize(file_name) == 4576820670:
            print("************** Uncompress already exists tar format data **************", flush=True)
            _unzip(file_name)
    if not _check_uncompressed_kaggle_display_advertising_files(dataset_path):
        _fetch_and_unzip_by_wget(remote_url, file_name)
    else:
        print("{} already have uncompressed kaggle display advertising dataset.".format(dataset_path), flush=True)

    return os.path.join(dataset_path)


download_checker = {
    'mnist': _download_mnist,
    'cifar10': _download_cifar10,
    'cifar100': _download_cifar100,
    'voc': _download_voc,
    'kaggle_display_advertising': _download_kaggle_display_advertising,
}


def download_dataset(dataset_name, local_path='.'):
    r'''
    This function is defined to easily download any public dataset
    without specifing much details.

    Args:
        dataset_name (str): The official name of dataset, currently supports `mnist`, `cifar10` and `cifar100`.
        local_path (str): Specifies the local location of dataset to be downloaded.
            Default: `.`.

    Returns:
        str, the source location of dataset downloaded.

    Examples:
        >>> from tinyms.data import download_dataset
        >>>
        >>> ds_path = download_dataset('mnist')
    '''
    download_func = download_checker.get(dataset_name)
    if download_func is None:
        print("Currently dataset_name only supports {}!".format(list(download_checker.keys())))
        sys.exit(0)

    return download_func(local_path)


def load_resized_img(path, width=256, height=256):
    """
    Load image with RGB and resize to (256, 256).

    Args:
        path (str): image path.
        width (int): image width, default: 256.
        height (int): image height, default: 256.

    Returns:
        PIL image class.
    """
    return Image.open(path).convert('RGB').resize((width, height))


def load_img(path):
    """
    Load image with RGB.

    Args:
        path (str): image path.

    Returns:
        PIL image class.
    """
    if path is None or not is_image(path):
        assert path, '%s is none or is not an image'
    return Image.open(path).convert('RGB')


def save_image(img, img_path):
    """
    Save a numpy image to the disk.

    Args:
        img (Union[numpy.ndarray, Tensor]): Image to save.
        img_path (str): The path of the image.
    """
    if isinstance(img, Tensor):
        # Decode a [1, C, H, W] Tensor to image numpy array.
        mean = 0.5 * 255
        std = 0.5 * 255
        img = (img.asnumpy()[0] * std + mean).astype(np.uint8).transpose((1, 2, 0))
    elif not isinstance(img, np.ndarray):
        raise ValueError("img should be Tensor or numpy array, but get {}".format(type(img)))
    img_pil = Image.fromarray(img)
    img_pil.save(img_path)



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

from ...callbacks import ModelCheckpoint, CheckpointConfig, LossTimeMonitor

__all__ = ['mobilenetv2_cb']


def mobilenetv2_cb(device_target, lr, is_saving_checkpoint, save_checkpoint_epochs, step_size):
    """
    Get mobilenetv2 callback.

    Args:
       device_target (str): 'CPU', 'GPU or 'Ascend'.
       lr (Tensor): learning rate.
       is_saving_checkpoint (bool): Whether to save the checkpoint file.
       save_checkpoint_epochs (int): Save checkpoint epochs.
       step_size (int): dataset.get_dataset_size().

    Returns:
       Callback list.
    """
    cb = None
    if device_target in ("CPU", "GPU"):
        cb = [LossTimeMonitor(lr_init=lr.asnumpy())]

        if is_saving_checkpoint:
            config_ck = CheckpointConfig(save_checkpoint_steps=save_checkpoint_epochs * step_size,
                                         keep_checkpoint_max=10)
            ckpt_save_dir = "./"
            ckpt_cb = ModelCheckpoint(prefix="mobilenetv2_cifar10", directory=ckpt_save_dir, config=config_ck)
            cb += [ckpt_cb]
    return cb



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""AdamWeightDecayForBert, a customized Adam for bert. Input: gradient, overflow flag."""
import numpy as np

from mindspore.nn.learning_rate_schedule import LearningRateSchedule, PolynomialDecayLR, WarmUpLR
from mindspore._checkparam import Validator as validator
from mindspore._checkparam import Rel


import tinyms as ts
from tinyms import Tensor
from tinyms import primitives as P
from tinyms import context
from . import AdamWeightDecay, Lamb, Momentum, thor, Optimizer


__all__ = ['AdamWeightDecayForBert', 'AdamWeightDecayOp']



_adam_opt = P.MultitypeFuncGraph("adam_opt")


def _check_param_value(beta1, beta2, eps, prim_name):
    """Check the type of inputs."""
    validator.check_value_type("beta1", beta1, [float], prim_name)
    validator.check_value_type("beta2", beta2, [float], prim_name)
    validator.check_value_type("eps", eps, [float], prim_name)
    validator.check_float_range(beta1, 0.0, 1.0, Rel.INC_NEITHER, "beta1", prim_name)
    validator.check_float_range(beta2, 0.0, 1.0, Rel.INC_NEITHER, "beta2", prim_name)
    validator.check_positive_float(eps, "eps", prim_name)

class AdamWeightDecayForBert(Optimizer):
    """
    Implements the Adam algorithm to fix the weight decay.

    Note:
        When separating parameter groups, the weight decay in each group will be applied on the parameters if the
        weight decay is positive. When not separating parameter groups, the `weight_decay` in the API will be applied
        on the parameters without 'beta' or 'gamma' in their names if `weight_decay` is positive.

        To improve parameter groups performance, the customized order of parameters can be supported.

    Args:
        params (Union[list[Parameter], list[dict]]): When the `params` is a list of `Parameter` which will be updated,
            the element in `params` must be class `Parameter`. When the `params` is a list of `dict`, the "params",
            "lr", "weight_decay" and "order_params" are the keys can be parsed.

            - params: Required. The value must be a list of `Parameter`.

            - lr: Optional. If "lr" is in the keys, the value of the corresponding learning rate will be used.
              If not, the `learning_rate` in the API will be used.

            - weight_decay: Optional. If "weight_decay" is in the keys, the value of the corresponding weight decay
              will be used. If not, the `weight_decay` in the API will be used.

            - order_params: Optional. If "order_params" is in the keys, the value must be the order of parameters and
              the order will be followed in the optimizer. There are no other keys in the `dict` and the parameters
              which in the 'order_params' must be in one of group parameters.

        learning_rate (Union[float, Tensor, Iterable, LearningRateSchedule]): A value or a graph for the learning rate.
            When the learning_rate is an Iterable or a Tensor in a 1D dimension, use the dynamic learning rate, then
            the i-th step will take the i-th value as the learning rate. When the learning_rate is LearningRateSchedule,
            use dynamic learning rate, the i-th learning rate will be calculated during the process of training
            according to the formula of LearningRateSchedule. When the learning_rate is a float or a Tensor in a zero
            dimension, use fixed learning rate. Other cases are not supported. The float learning rate must be
            equal to or greater than 0. If the type of `learning_rate` is int, it will be converted to float.
            Default: 1e-3.
        beta1 (float): The exponential decay rate for the 1st moment estimations. Default: 0.9.
            Should be in range (0.0, 1.0).
        beta2 (float): The exponential decay rate for the 2nd moment estimations. Default: 0.999.
            Should be in range (0.0, 1.0).
        eps (float): Term added to the denominator to improve numerical stability. Default: 1e-6.
            Should be greater than 0.
        weight_decay (float): Weight decay (L2 penalty). It must be equal to or greater than 0. Default: 0.0.

    Inputs:
        - **gradients** (tuple[Tensor]) - The gradients of `params`, the shape is the same as `params`.
        - **overflow** (tuple[Tensor]) - The overflow flag in dynamiclossscale.

    Outputs:
        tuple[bool], all elements are True.

    Supported Platforms:
        ``Ascend`` ``GPU``

    Examples:
        >>> net = Net()
        >>> #1) All parameters use the same learning rate and weight decay
        >>> optim = AdamWeightDecay(params=net.trainable_params())
        >>>
        >>> #2) Use parameter groups and set different values
        >>> conv_params = list(filter(lambda x: 'conv' in x.name, net.trainable_params()))
        >>> no_conv_params = list(filter(lambda x: 'conv' not in x.name, net.trainable_params()))
        >>> group_params = [{'params': conv_params, 'weight_decay': 0.01},
        ...                 {'params': no_conv_params, 'lr': 0.01},
        ...                 {'order_params': net.trainable_params()}]
        >>> optim = AdamWeightDecay(group_params, learning_rate=0.1, weight_decay=0.0)
        >>> # The conv_params's parameters will use default learning rate of 0.1 and weight decay of 0.01.
        >>> # The no_conv_params's parameters will use learning rate of 0.01 and default weight decay of 0.0.
        >>> # The final parameters order in which the optimizer will be followed is the value of 'order_params'.
        >>>
        >>> loss = layers.SoftmaxCrossEntropyWithLogits()
        >>> model = Model(net)
        >>> model.compile(loss_fn=loss, optimizer=optim)
   """
    def __init__(self, params, learning_rate=1e-3, beta1=0.9, beta2=0.999, eps=1e-6, weight_decay=0.0):
        super(AdamWeightDecayForBert, self).__init__(learning_rate, params, weight_decay)
        _check_param_value(beta1, beta2, eps, self.cls_name)
        self.beta1 = ts.array([beta1], dtype=ts.float32)
        self.beta2 = ts.array([beta2], dtype=ts.float32)
        self.eps = ts.array([eps], dtype=ts.float32)
        self.moments1 = self.parameters.clone(prefix="adam_m", init='zeros')
        self.moments2 = self.parameters.clone(prefix="adam_v", init='zeros')
        self.hyper_map = P.HyperMap()
        self.op_select = P.Select()
        self.op_cast = P.Cast()
        self.op_reshape = P.Reshape()
        self.op_shape = P.Shape()

    def construct(self, gradients, overflow):
        """AdamWeightDecayForBert"""
        lr = self.get_lr()
        cond = self.op_cast(P.Fill()(ts.int32, self.op_shape(self.beta1), 1) *\
                            self.op_reshape(overflow, (())), ts.bool_)
        beta1 = self.op_select(cond, self.op_cast(ts.array((1.0,)), ts.float32), self.beta1)
        beta2 = self.op_select(cond, self.op_cast(ts.array((1.0,)), ts.float32), self.beta2)
        if self.is_group:
            if self.is_group_lr:
                optim_result = self.hyper_map(P.Partial()(_adam_opt, self.beta1, self.beta2, self.eps),
                                              lr, self.weight_decay, self.parameters, self.moments1, self.moments2,
                                              gradients, self.decay_flags, self.optim_filter)
            else:
                optim_result = self.hyper_map(P.Partial()(_adam_opt, beta1, beta2, self.eps, lr, overflow),
                                              self.weight_decay, self.parameters, self.moments1, self.moments2,
                                              gradients, self.decay_flags, self.optim_filter)
        else:
            optim_result = self.hyper_map(P.Partial()(_adam_opt, self.beta1, self.beta2, self.eps, lr, self.weight_decay),
                                          self.parameters, self.moments1, self.moments2,
                                          gradients, self.decay_flags, self.optim_filter)
        if self.use_parallel:
            self.broadcast_params(optim_result)
        return optim_result

class AdamWeightDecayOp(Optimizer):
    """
    Implements the Adam algorithm to fix the weight decay. It is a complete operator, not a combination of other ops.

    Note:
        When separating parameter groups, the weight decay in each group will be applied on the parameters if the
        weight decay is positive. When not separating parameter groups, the `weight_decay` in the API will be applied
        on the parameters without 'beta' or 'gamma' in their names if `weight_decay` is positive.

        To improve parameter groups performance, the customized order of parameters can be supported.

    Args:
        params (Union[list[Parameter], list[dict]]): When the `params` is a list of `Parameter` which will be updated,
            the element in `params` must be class `Parameter`. When the `params` is a list of `dict`, the "params",
            "lr", "weight_decay" and "order_params" are the keys can be parsed.

            - params: Required. The value must be a list of `Parameter`.

            - lr: Optional. If "lr" is in the keys, the value of the corresponding learning rate will be used.
              If not, the `learning_rate` in the API will be used.

            - weight_decay: Optional. If "weight_decay" is in the keys, the value of the corresponding weight decay
              will be used. If not, the `weight_decay` in the API will be used.

            - order_params: Optional. If "order_params" is in the keys, the value must be the order of parameters and
              the order will be followed in the optimizer. There are no other keys in the `dict` and the parameters
              which in the 'order_params' must be in one of group parameters.

        learning_rate (Union[float, Tensor, Iterable, LearningRateSchedule]): A value or a graph for the learning rate.
            When the learning_rate is an Iterable or a Tensor in a 1D dimension, use the dynamic learning rate, then
            the i-th step will take the i-th value as the learning rate. When the learning_rate is LearningRateSchedule,
            use dynamic learning rate, the i-th learning rate will be calculated during the process of training
            according to the formula of LearningRateSchedule. When the learning_rate is a float or a Tensor in a zero
            dimension, use fixed learning rate. Other cases are not supported. The float learning rate must be
            equal to or greater than 0. If the type of `learning_rate` is int, it will be converted to float.
            Default: 1e-3.
        beta1 (float): The exponential decay rate for the 1st moment estimations. Default: 0.9.
            Should be in range (0.0, 1.0).
        beta2 (float): The exponential decay rate for the 2nd moment estimations. Default: 0.999.
            Should be in range (0.0, 1.0).
        eps (float): Term added to the denominator to improve numerical stability. Default: 1e-6.
            Should be greater than 0.
        weight_decay (float): Weight decay (L2 penalty). It must be equal to or greater than 0. Default: 0.0.

    Inputs:
        - **gradients** (tuple[Tensor]) - The gradients of `params`, the shape is the same as `params`.

    Outputs:
        tuple[bool], all elements are True.

    Supported Platforms:
        ``GPU``

    Examples:
        >>> net = Net()
        >>> #1) All parameters use the same learning rate and weight decay
        >>> optim = AdamWeightDecayOp(params=net.trainable_params())
        >>>
        >>> #2) Use parameter groups and set different values
        >>> conv_params = list(filter(lambda x: 'conv' in x.name, net.trainable_params()))
        >>> no_conv_params = list(filter(lambda x: 'conv' not in x.name, net.trainable_params()))
        >>> group_params = [{'params': conv_params, 'weight_decay': 0.01},
        ...                 {'params': no_conv_params, 'lr': 0.01},
        ...                 {'order_params': net.trainable_params()}]
        >>> optim = AdamWeightDecayOp(group_params, learning_rate=0.1, weight_decay=0.0)
        >>> # The conv_params's parameters will use default learning rate of 0.1 and weight decay of 0.01.
        >>> # The no_conv_params's parameters will use learning rate of 0.01 and default weight decay of 0.0.
        >>> # The final parameters order in which the optimizer will be followed is the value of 'order_params'.
        >>>
        >>> loss = layers.SoftmaxCrossEntropyWithLogits()
        >>> model = Model(net)
        >>> model.compile(loss_fn=loss, optimizer=optim)
   """
    def __init__(self, params, learning_rate=1e-3, beta1=0.9, beta2=0.999, eps=1e-6, weight_decay=0.0):
        super(AdamWeightDecayOp, self).__init__(learning_rate, params, weight_decay)
        _check_param_value(beta1, beta2, eps, self.cls_name)
        self.beta1 = ts.array([beta1], dtype=ts.float32)
        self.beta2 = ts.array([beta2], dtype=ts.float32)
        self.eps = ts.array([eps], dtype=ts.float32)
        self.moments1 = self.parameters.clone(prefix="adam_m", init='zeros')
        self.moments2 = self.parameters.clone(prefix="adam_v", init='zeros')
        self.hyper_map = P.HyperMap()

    def construct(self, gradients):
        """AdamWeightDecayOp"""
        lr = self.get_lr()
        if self.is_group:
            if self.is_group_lr:
                optim_result = self.hyper_map(P.Partial()(_adam_opt, self.beta1, self.beta2, self.eps),
                                              lr, self.weight_decay, self.parameters, self.moments1, self.moments2,
                                              gradients, self.decay_flags, self.optim_filter)
            else:
                optim_result = self.hyper_map(P.Partial()(_adam_opt, self.beta1, self.beta2, self.eps, lr),
                                              self.weight_decay, self.parameters, self.moments1, self.moments2,
                                              gradients, self.decay_flags, self.optim_filter)
        else:
            optim_result = self.hyper_map(P.Partial()(_adam_opt, self.beta1, self.beta2, self.eps, lr, self.weight_decay),
                                          self.parameters, self.moments1, self.moments2,
                                          gradients, self.decay_flags, self.optim_filter)
        if self.use_parallel:
            self.broadcast_params(optim_result)
        return optim_result




class BertLearningRate(LearningRateSchedule):
    """
    Warmup-decay learning rate for Bert network.
    """
    def __init__(self, learning_rate, end_learning_rate, warmup_steps, decay_steps, power):
        super(BertLearningRate, self).__init__()
        self.warmup_flag = False
        if warmup_steps > 0:
            self.warmup_flag = True
            self.warmup_lr = WarmUpLR(learning_rate, warmup_steps)
        self.decay_lr = PolynomialDecayLR(learning_rate, end_learning_rate, decay_steps, power)
        self.warmup_steps = ts.array([warmup_steps], dtype=ts.float32)

        self.greater = P.Greater()
        self.one = ts.array([1.0], dtype=ts.float32)
        self.cast = P.Cast()

    def construct(self, global_step):
        decay_lr = self.decay_lr(global_step)
        if self.warmup_flag:
            is_warmup = self.cast(self.greater(self.warmup_steps, global_step), ts.float32)
            warmup_lr = self.warmup_lr(global_step)
            lr = (self.one - is_warmup) * decay_lr + is_warmup * warmup_lr
        else:
            lr = decay_lr
        return lr



def _get_poly_lr(global_step, lr_init, lr_end, lr_max, warmup_steps, total_steps, poly_power):
    """
    generate learning rate array

    Args:
       global_step(int): current step
       lr_init(float): init learning rate
       lr_end(float): end learning rate
       lr_max(float): max learning rate
       warmup_steps(int): number of warmup epochs
       total_steps(int): total epoch of training
       poly_power(int): poly learning rate power

    Returns:
       np.array, learning rate array
    """
    lr_each_step = []
    if warmup_steps != 0:
        inc_each_step = (float(lr_max) - float(lr_init)) / float(warmup_steps)
    else:
        inc_each_step = 0
    for i in range(total_steps):
        if i < warmup_steps:
            lr = float(lr_init) + inc_each_step * float(i)
        else:
            base = (1.0 - (float(i) - float(warmup_steps)) / (float(total_steps) - float(warmup_steps)))
            lr = float(lr_max - lr_end) * (base ** poly_power)
            lr = lr + lr_end
            if lr < 0.0:
                lr = 0.0
        lr_each_step.append(lr)

    learning_rate = np.array(lr_each_step).astype(np.float32)
    current_step = global_step
    learning_rate = learning_rate[current_step:]
    return learning_rate



def get_bert_thor_damping(damping_max=5e-2, damping_min=1e-6, damping_power=1.0, damping_total_steps=30000):
    damping = _get_poly_lr(global_step=0, lr_init=0.0, lr_end=damping_min, lr_max=damping_max, warmup_steps=0,
                           total_steps=damping_total_steps, poly_power=damping_power)
    return Tensor(damping)


def get_bert_thor_lr(lr_max=0.0034, lr_min=3.244e-05, lr_power=1.0, lr_total_steps=30000):
    learning_rate = _get_poly_lr(global_step=0, lr_init=0.0, lr_end=lr_min, lr_max=lr_max, warmup_steps=0,
                                 total_steps=lr_total_steps, poly_power=lr_power)
    return Tensor(learning_rate)





def get_optimizer(args_opt, network, cfg, bert_net_cfg):
    """get bert optimizer, support Lamb, Momentum, AdamWeightDecay."""
    if cfg.optimizer == 'Lamb':
        lr_schedule = BertLearningRate(learning_rate=cfg.Lamb.learning_rate,
                                       end_learning_rate=cfg.Lamb.end_learning_rate,
                                       warmup_steps=cfg.Lamb.warmup_steps,
                                       decay_steps=args_opt.train_steps,
                                       power=cfg.Lamb.power)
        params = network.trainable_params()
        decay_params = list(filter(cfg.Lamb.decay_filter, params))
        other_params = list(filter(lambda x: not cfg.Lamb.decay_filter(x), params))
        group_params = [{'params': decay_params, 'weight_decay': cfg.Lamb.weight_decay},
                        {'params': other_params},
                        {'order_params': params}]
        optimizer = Lamb(group_params, learning_rate=lr_schedule, eps=cfg.Lamb.eps)
    elif cfg.optimizer == 'Momentum':
        optimizer = Momentum(network.trainable_params(), learning_rate=cfg.Momentum.learning_rate,
                             momentum=cfg.Momentum.momentum)
    elif cfg.optimizer == 'AdamWeightDecay':
        lr_schedule = BertLearningRate(learning_rate=cfg.AdamWeightDecay.learning_rate,
                                       end_learning_rate=cfg.AdamWeightDecay.end_learning_rate,
                                       warmup_steps=cfg.AdamWeightDecay.warmup_steps,
                                       decay_steps=args_opt.train_steps,
                                       power=cfg.AdamWeightDecay.power)
        params = network.trainable_params()
        decay_params = list(filter(cfg.AdamWeightDecay.decay_filter, params))
        other_params = list(filter(lambda x: not cfg.AdamWeightDecay.decay_filter(x), params))
        group_params = [{'params': decay_params, 'weight_decay': cfg.AdamWeightDecay.weight_decay},
                        {'params': other_params, 'weight_decay': 0.0},
                        {'order_params': params}]

        if args_opt.enable_lossscale == "true" and args_opt.device_target == 'GPU':
            optimizer = AdamWeightDecayForBert(group_params, learning_rate=lr_schedule, eps=cfg.AdamWeightDecay.eps)
        elif context.get_context("mode") == context.PYNATIVE_MODE and args_opt.device_target == 'GPU':
            optimizer = AdamWeightDecayOp(group_params, learning_rate=lr_schedule, eps=cfg.AdamWeightDecay.eps)
        else:
            optimizer = AdamWeightDecay(group_params, learning_rate=lr_schedule, eps=cfg.AdamWeightDecay.eps)

    elif cfg.optimizer == "Thor":

        lr = get_bert_thor_lr(cfg.Thor.lr_max, cfg.Thor.lr_min, cfg.Thor.lr_power, cfg.Thor.lr_total_steps)
        damping = get_bert_thor_damping(cfg.Thor.damping_max, cfg.Thor.damping_min, cfg.Thor.damping_power,
                                        cfg.Thor.damping_total_steps)
        split_indices = None

        if bert_net_cfg.num_hidden_layers == 12:
            if bert_net_cfg.use_relative_positions:
                split_indices = [29, 58, 87, 116, 145, 174, 203, 217]
            else:
                split_indices = [28, 55, 82, 109, 136, 163, 190, 205]
        elif bert_net_cfg.num_hidden_layers == 24:
            if bert_net_cfg.use_relative_positions:
                split_indices = [30, 90, 150, 210, 270, 330, 390, 421]
            else:
                split_indices = [38, 93, 148, 203, 258, 313, 368, 397]

        optimizer = thor(network, lr, damping, cfg.Thor.momentum,
                         cfg.Thor.weight_decay, cfg.Thor.loss_scale, cfg.batch_size,
                         decay_filter=lambda x: 'layernorm' not in x.name.lower() and 'bias' not in x.name.lower(),
                         split_indices=split_indices)
    else:
        raise ValueError("Don't support optimizer {}, only support [Lamb, Momentum, AdamWeightDecay, Thor]".
                         format(cfg.optimizer))

    return optimizer


# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===========================================================================
'''
Bert for finetune script.
'''

import tinyms as ts
from tinyms import Tensor
from tinyms import Parameter
from tinyms import context
from tinyms import layers
from tinyms import primitives as P
from tinyms.initializers import  initializer

from .bert_for_pretraining import clip_grad
from .finetune_eval_model import BertCLSModel, BertNERModel, BertSquadModel


GRADIENT_CLIP_TYPE = 1
GRADIENT_CLIP_VALUE = 1.0
grad_scale = P.MultitypeFuncGraph("grad_scale")
reciprocal = P.Reciprocal()
@grad_scale.register("Tensor", "Tensor")
def tensor_grad_scale(scale, grad):
    return grad * reciprocal(scale)

_grad_overflow = P.MultitypeFuncGraph("_grad_overflow")
grad_overflow = P.FloatStatus()
@_grad_overflow.register("Tensor")
def _tensor_grad_overflow(grad):
    return grad_overflow(grad)

class CrossEntropyCalculation(layers.Layer):
    """
    Cross Entropy loss
    """

    def __init__(self, is_training=True):
        super(CrossEntropyCalculation, self).__init__()
        self.onehot = P.OneHot()
        self.on_value = Tensor(1.0, ts.float32)
        self.off_value = Tensor(0.0, ts.float32)
        self.reduce_sum = P.ReduceSum()
        self.reduce_mean = P.ReduceMean()
        self.reshape = P.Reshape()
        self.last_idx = (-1,)
        self.neg = P.Neg()
        self.cast = P.Cast()
        self.is_training = is_training

    def construct(self, logits, label_ids, num_labels):
        if self.is_training:
            label_ids = self.reshape(label_ids, self.last_idx)
            one_hot_labels = self.onehot(label_ids, num_labels, self.on_value, self.off_value)
            per_example_loss = self.neg(self.reduce_sum(one_hot_labels * logits, self.last_idx))
            loss = self.reduce_mean(per_example_loss, self.last_idx)
            return_value = self.cast(loss, ts.float32)
        else:
            return_value = logits * 1.0
        return return_value


class BertFinetuneLayer(layers.Layer):
    """
    Especifically defined for finetuning where only four inputs tensor are needed.
    """

    def __init__(self, network, optimizer, scale_update_layer=None):

        super(BertFinetuneLayer, self).__init__(auto_prefix=False)
        self.network = network
        self.network.set_grad()
        self.weights = optimizer.parameters
        self.optimizer = optimizer
        self.optimizer.global_step = Parameter(initializer(0., [1,]), name='global_step')
        self.grad = P.GradOperation(get_by_list=True,
                                    sens_param=True)
        self.allreduce = P.AllReduce()
        self.grad_reducer = None
        self.cast = P.Cast()
        self.gpu_target = False
        if context.get_context("device_target") == "GPU":
            self.gpu_target = True
            self.float_status = P.FloatStatus()
            self.addn = P.AddN()
            self.reshape = P.Reshape()
        else:
            self.alloc_status = P.NPUAllocFloatStatus()
            self.get_status = P.NPUGetFloatStatus()
            self.clear_before_grad = P.NPUClearFloatStatus()
        self.reduce_sum = P.ReduceSum(keep_dims=False)
        self.depend_parameter_use = P.Depend()
        self.base = Tensor(1, ts.float32)
        self.less_equal = P.LessEqual()
        self.hyper_map = P.HyperMap()
        self.loss_scale = None
        self.loss_scaling_manager = scale_update_layer
        if scale_update_layer:
            self.loss_scale = Parameter(Tensor(scale_update_layer.get_loss_scale(), dtype=ts.float32),
                                        name="loss_scale")

    def construct(self,
                  input_ids,
                  input_mask,
                  token_type_id,
                  label_ids,
                  sens=None):
        """Bert Finetune"""

        weights = self.weights
        init = False
        loss = self.network(input_ids,
                            input_mask,
                            token_type_id,
                            label_ids)
        if sens is None:
            scaling_sens = self.loss_scale
        else:
            scaling_sens = sens

        if not self.gpu_target:
            init = self.alloc_status()
            clear_before_grad = self.clear_before_grad(init)
            loss = P.depend(loss, init)
            self.depend_parameter_use(clear_before_grad, scaling_sens)
        grads = self.grad(self.network, weights)(input_ids,
                                                 input_mask,
                                                 token_type_id,
                                                 label_ids,
                                                 self.cast(scaling_sens,
                                                           ts.float32))
        grads = self.hyper_map(P.partial(grad_scale, scaling_sens), grads)
        grads = self.hyper_map(P.partial(clip_grad, GRADIENT_CLIP_TYPE, GRADIENT_CLIP_VALUE), grads)
        if not self.gpu_target:
            flag = self.get_status(init)
            flag_sum = self.reduce_sum(init, (0,))
            grads = P.depend(grads, flag)
            flag_sum = P.depend(flag_sum, flag)
        else:
            flag_sum = self.hyper_map(P.partial(_grad_overflow), grads)
            flag_sum = self.addn(flag_sum)
            flag_sum = self.reshape(flag_sum, (()))

        cond = self.less_equal(self.base, flag_sum)
        overflow = cond
        if sens is None:
            overflow = self.loss_scaling_manager(self.loss_scale, cond)
        if overflow:
            succ = False
        else:
            succ = self.optimizer(grads)
        ret = (loss, cond)
        return P.depend(ret, succ)


class BertSquadLayer(layers.Layer):
    """
    specifically defined for finetuning where only four inputs tensor are needed.
    """

    def __init__(self, network, optimizer, scale_update_layer=None):
        super(BertSquadLayer, self).__init__(auto_prefix=False)
        self.network = network
        self.network.set_grad()
        self.weights = optimizer.parameters
        self.optimizer = optimizer
        self.grad = P.GradOperation(get_by_list=True, sens_param=True)
        self.allreduce = P.AllReduce()
        self.grad_reducer = None
        self.cast = P.Cast()
        self.alloc_status = P.NPUAllocFloatStatus()
        self.get_status = P.NPUGetFloatStatus()
        self.clear_before_grad = P.NPUClearFloatStatus()
        self.reduce_sum = P.ReduceSum(keep_dims=False)
        self.depend_parameter_use = P.Depend()
        self.base = Tensor(1, ts.float32)
        self.less_equal = P.LessEqual()
        self.hyper_map = P.HyperMap()
        self.loss_scale = None
        self.loss_scaling_manager = scale_update_layer
        if scale_update_layer:
            self.loss_scale = Parameter(Tensor(scale_update_layer.get_loss_scale(), dtype=ts.float32),
                                        name="loss_scale")
    def construct(self,
                  input_ids,
                  input_mask,
                  token_type_id,
                  start_position,
                  end_position,
                  unique_id,
                  is_impossible,
                  sens=None):
        """BertSquad"""
        weights = self.weights
        init = self.alloc_status()
        loss = self.network(input_ids,
                            input_mask,
                            token_type_id,
                            start_position,
                            end_position,
                            unique_id,
                            is_impossible)
        if sens is None:
            scaling_sens = self.loss_scale
        else:
            scaling_sens = sens
        grads = self.grad(self.network, weights)(input_ids,
                                                 input_mask,
                                                 token_type_id,
                                                 start_position,
                                                 end_position,
                                                 unique_id,
                                                 is_impossible,
                                                 self.cast(scaling_sens,
                                                           ts.float32))
        clear_before_grad = self.clear_before_grad(init)
        loss = P.depend(loss, init)
        self.depend_parameter_use(clear_before_grad, scaling_sens)
        grads = self.hyper_map(P.partial(grad_scale, scaling_sens), grads)
        grads = self.hyper_map(P.partial(clip_grad, GRADIENT_CLIP_TYPE, GRADIENT_CLIP_VALUE), grads)
        flag = self.get_status(init)
        flag_sum = self.reduce_sum(init, (0,))
        cond = self.less_equal(self.base, flag_sum)
        P.depend(grads, flag)
        P.depend(flag, flag_sum)
        overflow = cond
        if sens is None:
            overflow = self.loss_scaling_manager(self.loss_scale, cond)
        if overflow:
            succ = False
        else:
            succ = self.optimizer(grads)
        ret = (loss, cond)
        return P.depend(ret, succ)


class BertCLS(layers.Layer):
    """
    Train interface for classification finetuning task.
    """

    def __init__(self, config, is_training, num_labels=2, dropout_prob=0.0, use_one_hot_embeddings=False,
                 assessment_method=""):
        super(BertCLS, self).__init__()
        self.bert = BertCLSModel(config, is_training, num_labels, dropout_prob, use_one_hot_embeddings,
                                 assessment_method)
        self.loss = CrossEntropyCalculation(is_training)
        self.num_labels = num_labels
        self.assessment_method = assessment_method
        self.is_training = is_training
    def construct(self, input_ids, input_mask, token_type_id, label_ids):
        logits = self.bert(input_ids, input_mask, token_type_id)
        if self.assessment_method == "spearman_correlation":
            if self.is_training:
                loss = self.loss(logits, label_ids)
            else:
                loss = logits
        else:
            loss = self.loss(logits, label_ids, self.num_labels)
        return loss


class BertNER(layers.Layer):
    """
    Train interface for sequence labeling finetuning task.
    """

    def __init__(self, config, batch_size, is_training, num_labels=11, use_crf=False,
                 tag_to_index=None, dropout_prob=0.0, use_one_hot_embeddings=False):
        super(BertNER, self).__init__()
        self.bert = BertNERModel(config, is_training, num_labels, use_crf, dropout_prob, use_one_hot_embeddings)
        if use_crf:
            if not tag_to_index:
                raise Exception("The dict for tag-index mapping should be provided for CRF.")
            from src.CRF import CRF
            self.loss = CRF(tag_to_index, batch_size, config.seq_length, is_training)
        else:
            self.loss = CrossEntropyCalculation(is_training)
        self.num_labels = num_labels
        self.use_crf = use_crf
    def construct(self, input_ids, input_mask, token_type_id, label_ids):
        logits = self.bert(input_ids, input_mask, token_type_id)
        if self.use_crf:
            loss = self.loss(logits, label_ids)
        else:
            loss = self.loss(logits, label_ids, self.num_labels)
        return loss


class BertSquad(layers.Layer):
    '''
    Train interface for SQuAD finetuning task.
    '''

    def __init__(self, config, is_training, num_labels=2, dropout_prob=0.0, use_one_hot_embeddings=False):
        super(BertSquad, self).__init__()
        self.bert = BertSquadModel(config, is_training, num_labels, dropout_prob, use_one_hot_embeddings)
        self.loss = CrossEntropyCalculation(is_training)
        self.num_labels = num_labels
        self.seq_length = config.seq_length
        self.is_training = is_training
        self.total_num = Parameter(Tensor([0], ts.float32), name='total_num')
        self.start_num = Parameter(Tensor([0], ts.float32), name='start_num')
        self.end_num = Parameter(Tensor([0], ts.float32), name='end_num')
        self.sum = P.ReduceSum()
        self.equal = P.Equal()
        self.argmax = P.ArgMaxWithValue(axis=1)
        self.squeeze = P.Squeeze(axis=-1)

    def construct(self, input_ids, input_mask, token_type_id, start_position, end_position, unique_id, is_impossible):
        """interface for SQuAD finetuning task"""
        logits = self.bert(input_ids, input_mask, token_type_id)
        if self.is_training:
            unstacked_logits_0 = self.squeeze(logits[:, :, 0:1])
            unstacked_logits_1 = self.squeeze(logits[:, :, 1:2])
            start_loss = self.loss(unstacked_logits_0, start_position, self.seq_length)
            end_loss = self.loss(unstacked_logits_1, end_position, self.seq_length)
            total_loss = (start_loss + end_loss) / 2.0
        else:
            start_logits = self.squeeze(logits[:, :, 0:1])
            end_logits = self.squeeze(logits[:, :, 1:2])
            total_loss = (unique_id, start_logits, end_logits)
        return total_loss
